# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************


import FreeCAD
import FreeCADGui
import ArchBuildingPart

translate = FreeCAD.Qt.translate


class ifc_vp_object:
    """Base class for all blenderbim view providers"""

    def attach(self, vobj):
        self.Object = vobj.Object

    def getDisplayModes(self, obj):
        return []

    def getDefaultDisplayMode(self):
        return "FlatLines"

    def setDisplayMode(self, mode):
        return mode

    def onChanged(self, vobj, prop):
        if prop == "Visibility":
            for child in vobj.Object.Group:
                child.ViewObject.Visibility = vobj.Visibility
            return True
        elif prop == "LineColor" and vobj.Object.ShapeMode == "Coin":
            lc = vobj.LineColor
            basenode = vobj.RootNode.getChild(2).getChild(0)
            if basenode.getNumChildren() == 5:
                basenode[4][0][3].diffuseColor.setValue(lc[0], lc[1], lc[2])
        elif prop == "LineWidth" and vobj.Object.ShapeMode == "Coin":
            basenode = vobj.RootNode.getChild(2).getChild(0)
            if basenode.getNumChildren() == 5:
                basenode[4][0][4].lineWidth = vobj.LineWidth

    def __getstate__(self):
        return None

    def __setstate__(self, state):
        return None

    def updateData(self, obj, prop):
        if prop == "Shape" and getattr(obj, "Group", None):
            colors = []
            for child in obj.Group:
                if hasattr(child.ViewObject, "DiffuseColor"):
                    colors.extend(child.ViewObject.DiffuseColor)
            if colors:
                obj.ViewObject.DiffuseColor = colors

    def getIcon(self):
        return get_icon(self)

    def claimChildren(self):
        if hasattr(self.Object, "Group"):
            return self.Object.Group
        return []

    def setupContextMenu(self, vobj, menu):
        from . import ifc_tools  # lazy import
        from . import ifc_psets
        from . import ifc_materials
        from . import ifc_types
        from PySide import QtGui  # lazy import

        if FreeCADGui.activeWorkbench().name() != "BIMWorkbench":
            return

        icon = QtGui.QIcon(":/icons/IFC.svg")
        element = ifc_tools.get_ifc_element(vobj.Object)
        ifc_menu = None

        # IFC actions
        actions = []
        if element.is_a("IfcSpatialElement"):
            if FreeCADGui.ActiveDocument.ActiveView.getActiveObject("NativeIFC") == vobj.Object:
                t = translate("BIM", "Deactivate Container")
            else:
                t = translate("BIM", "Make Active Container")
            action_activate = QtGui.QAction(icon, t, menu)
            action_activate.triggered.connect(self.activate)
            menu.addAction(action_activate)
        if self.hasChildren(vobj.Object):
            action_expand = QtGui.QAction(icon, translate("BIM", "Expand Children"), menu)
            action_expand.triggered.connect(self.expandChildren)
            actions.append(action_expand)
        if vobj.Object.Group:
            action_shrink = QtGui.QAction(icon, translate("BIM", "Collapse Children"), menu)
            action_shrink.triggered.connect(self.collapseChildren)
            actions.append(action_shrink)
        if vobj.Object.ShapeMode == "Shape":
            t = translate("BIM", "Remove Shape")
        else:
            t = translate("BIM", "Load Shape")
        action_shape = QtGui.QAction(icon, t, menu)
        action_shape.triggered.connect(self.switchShape)
        actions.append(action_shape)
        if vobj.Object.ShapeMode == "None":
            action_coin = QtGui.QAction(icon, translate("BIM", "Load Representation"), menu)
            action_coin.triggered.connect(self.switchCoin)
            actions.append(action_coin)
        if element and ifc_tools.has_representation(element):
            action_geom = QtGui.QAction(icon, translate("BIM", "Add Geometry Properties"), menu)
            action_geom.triggered.connect(self.addGeometryProperties)
            actions.append(action_geom)
        action_tree = QtGui.QAction(icon, translate("BIM", "Show Geometry Tree"), menu)
        action_tree.triggered.connect(self.showTree)
        actions.append(action_tree)
        if ifc_psets.has_psets(self.Object):
            action_props = QtGui.QAction(icon, translate("BIM", "Expand Property Sets"), menu)
            action_props.triggered.connect(self.showProps)
            actions.append(action_props)
        if ifc_materials.get_material(self.Object):
            action_material = QtGui.QAction(icon, translate("BIM", "Load Material"), menu)
            action_material.triggered.connect(self.addMaterial)
            actions.append(action_material)
        if ifc_types.is_typable(self.Object):
            action_type = QtGui.QAction(icon, translate("BIM", "Convert to Type"), menu)
            action_type.triggered.connect(self.convertToType)
            actions.append(action_type)
        if actions:
            ifc_menu = QtGui.QMenu("IFC")
            ifc_menu.setIcon(icon)
            ifc_menu.addActions(actions)
            menu.addMenu(ifc_menu)

        # generic actions
        ficon = QtGui.QIcon.fromTheme("folder", QtGui.QIcon(":/icons/folder.svg"))
        action_group = QtGui.QAction(ficon, translate("BIM", "New Group"), menu)
        action_group.triggered.connect(self.createGroup)
        menu.addAction(action_group)

        # return submenu for derivated classes
        return ifc_menu

    def hasChildren(self, obj):
        """Returns True if this IFC object can be decomposed"""

        from . import ifc_tools  # lazy import

        ifcfile = ifc_tools.get_ifcfile(obj)
        if ifcfile:
            return ifc_tools.can_expand(obj, ifcfile)
        return False

    def expandChildren(self, obj=None):
        """Creates children of this object"""

        from . import ifc_tools  # lazy import
        from PySide import QtCore, QtGui

        if not obj:
            obj = self.Object
        ifcfile = ifc_tools.get_ifcfile(obj)
        nc = []
        if ifcfile:
            nc = ifc_tools.create_children(
                obj, ifcfile, recursive=False, assemblies=True, expand=False
            )
        obj.Document.recompute()
        FreeCADGui.updateGui()

        # expand the item in the tree view
        mw = FreeCADGui.getMainWindow()
        tree = mw.findChild(QtGui.QDockWidget, "Model")
        model = tree.findChild(QtGui.QWidget, "Model")
        splitter = model.findChild(QtGui.QSplitter)
        if splitter and len(splitter.children()) > 1:
            if splitter.children()[1].children():
                tree = splitter.children()[1].children()[0]
                it = tree.findItems(obj.Label, QtCore.Qt.MatchRecursive, 0)
                if it:
                    it[0].setExpanded(True)
                    for i in range(it[0].childCount()):
                        it[0].child(i).setExpanded(True)
        return nc

    def collapseChildren(self):
        """Collapses the children of this object"""

        objs = self.Object.Group
        for o in objs:
            objs.extend(self.getOwnChildren(o))
        for o in objs:
            if hasattr(o, "Proxy"):
                # this prevents to trigger the deletion inside the IFC file
                o.Proxy.nodelete = True
        names = [o.Name for o in objs]
        for name in names:
            self.Object.Document.removeObject(name)
        self.Object.Document.recompute()

    def getOwnChildren(self, obj):
        """Recursively gets the children only used by this object"""
        children = []
        for child in obj.OutList:
            if len(child.InList) == 1 and child.InList[0] == obj:
                children.append(child)
                children.extend(self.getOwnChildren(child))
        return children

    def switchShape(self):
        """Switch this object between shape and coin"""

        if self.Object.ShapeMode == "Shape":
            self.Object.ShapeMode = "Coin"
            import Part  # lazy loading

            self.Object.Shape = Part.Shape()
        else:
            self.Object.ShapeMode = "Shape"
        self.Object.Document.recompute()
        self.Object.ViewObject.DiffuseColor = self.Object.ViewObject.DiffuseColor
        self.Object.ViewObject.signalChangeIcon()

    def switchCoin(self):
        """Switch this object between coin and no representation"""

        changed = []
        if self.Object.ShapeMode == "None":
            self.Object.ShapeMode = "Coin"
            changed.append(self.Object.ViewObject)
        # reveal children
        for child in self.Object.OutListRecursive:
            if getattr(child, "ShapeMode", 0) == 2:
                child.ShapeMode = 1
                changed.append(child.ViewObject)
        self.Object.Document.recompute()
        for vobj in changed:
            vobj.DiffuseColor = vobj.DiffuseColor

    def addGeometryProperties(self):
        """Adds geometry properties to this object"""

        from . import ifc_geometry  # lazy loading

        ifc_geometry.add_geom_properties(self.Object)

    def addMaterial(self):
        """Adds a material to this object"""

        from . import ifc_materials  # lazy loading

        ifc_materials.show_material(self.Object)
        self.Object.Document.recompute()

    def showTree(self):
        """Shows a dialog with a geometry tree for the object"""

        from . import ifc_tools  # lazy loading
        from . import ifc_tree  # lazy loading

        element = ifc_tools.get_ifc_element(self.Object)
        if element:
            ifc_tree.show_geometry_tree(element)

    def showProps(self):
        """Expands property sets"""

        from . import ifc_psets  # lazy loading

        ifc_psets.show_psets(self.Object)
        self.Object.Document.recompute()

    def canDragObjects(self):
        """Whether children can be removed by d&d"""

        return True

    def canDropObjects(self):
        """Whether objects can be added here by d&d or drop only"""

        return True

    def canDragObject(self, dragged_object):
        """Whether the given object can be removed by d&d"""

        return True

    def canDropObject(self, incoming_object):
        """Whether the object can be dropped here by d&d or drop only"""

        return True  # in principle, any object can be dropped and become IFC

    def dragObject(self, vobj, dragged_object):
        """Remove a child from the view provider by d&d"""

        from . import ifc_tools  # lazy import

        parent = vobj.Object
        ifc_tools.deaggregate(dragged_object, parent)

    def dropObject(self, vobj, incoming_object):
        """Add an object to the view provider by d&d"""

        from PySide import QtCore  # lazy import

        # delay the action to prevent the object to be deleted
        # before the end of the drop
        QtCore.QTimer.singleShot(100, lambda: self.onDrop(incoming_object))

    def onDrop(self, incoming_object):
        """Delayed action to be taken when dropping an object"""

        from . import ifc_tools  # lazy import

        ifc_tools.aggregate(incoming_object, self.Object)
        if self.hasChildren(self.Object):
            self.expandChildren(self.Object)

    def activate(self):
        """Marks this container as active"""

        if FreeCADGui.ActiveDocument.ActiveView.getActiveObject("NativeIFC") == self.Object:
            FreeCADGui.ActiveDocument.ActiveView.setActiveObject("NativeIFC", None)
        else:
            FreeCADGui.ActiveDocument.ActiveView.setActiveObject("NativeIFC", self.Object)

    def createGroup(self):
        """Creates a group under this object"""

        from . import ifc_tools  # lazy import

        group = self.Object.Document.addObject("App::DocumentObjectGroup", "Group")
        ifc_tools.aggregate(group, self.Object)
        self.Object.Document.recompute()

    def doubleClicked(self, vobj):
        """On double-click"""

        self.expandProperties(vobj)

    def expandProperties(self, vobj):
        """Expands everything that needs to be expanded"""

        from . import ifc_geometry  # lazy import
        from . import ifc_tools  # lazy import
        from . import ifc_psets  # lazy import
        from . import ifc_materials  # lazy import
        from . import ifc_layers  # lazy import
        from . import ifc_types  # lazy import
        from . import ifc_classification  # lazy import

        # generic data loading
        ifc_geometry.add_geom_properties(vobj.Object)
        ifc_psets.show_psets(vobj.Object)
        ifc_materials.show_material(vobj.Object)
        ifc_layers.add_layers(vobj.Object)
        ifc_types.show_type(vobj.Object)
        ifc_classification.show_classification(vobj.Object)

        # expand children
        if self.hasChildren(vobj.Object):
            self.expandChildren()
            return True

        # load shape
        element = ifc_tools.get_ifc_element(vobj.Object)
        if ifc_tools.has_representation(element):
            if vobj.Object.ShapeMode != "Shape":
                vobj.Object.ShapeMode = "Shape"
                vobj.Object.Document.recompute()
                return True
        return None

    def convertToType(self):
        """Converts this object to a type"""

        if not hasattr(self, "Object"):
            return
        from . import ifc_types

        ifc_types.convert_to_type(self.Object)
        self.Object.Document.recompute()


class ifc_vp_document(ifc_vp_object):
    """View provider for the IFC document object"""

    def getIcon(self):
        iconpath = ":/icons/IFC_document.svg"
        if self.Object.Modified:
            if not hasattr(self, "modicon"):
                self.modicon = overlay(iconpath, ":/icons/media-record.svg")
            return self.modicon
        else:
            return iconpath

    def setupContextMenu(self, vobj, menu):

        from PySide import QtGui  # lazy import

        if FreeCADGui.activeWorkbench().name() != "BIMWorkbench":
            return

        ifc_menu = super().setupContextMenu(vobj, menu)
        if not ifc_menu:
            ifc_menu = menu

        icon = QtGui.QIcon(":/icons/IFC.svg")
        if vobj.Object.Modified:
            action_diff = QtGui.QAction(icon, translate("BIM", "View Diff"), menu)
            action_diff.triggered.connect(self.diff)
            ifc_menu.addAction(action_diff)
            if vobj.Object.IfcFilePath:
                action_save = QtGui.QAction(icon, translate("BIM", "Save IFC File"), menu)
                action_save.triggered.connect(self.save)
                ifc_menu.addAction(action_save)
        action_saveas = QtGui.QAction(icon, translate("BIM", "Save IFC File As…"), menu)
        action_saveas.triggered.connect(self.saveas)
        ifc_menu.addAction(action_saveas)

    def save(self):
        """Saves the associated IFC file"""

        from . import ifc_tools  # lazy import

        ifc_tools.save(self.Object)
        self.Object.Document.recompute()

    def saveas(self):
        """Saves the associated IFC file to another file"""

        from . import ifc_tools  # lazy import

        sf = get_filepath(self.Object)
        if sf:
            ifc_tools.save(self.Object)
            self.replace_file(self.Object, sf)
            self.Object.Document.recompute()

    def replace_file(self, obj, newfile):
        """Asks the user if the attached file path needs to be replaced"""

        from PySide import QtGui  # lazy import

        msg = "Replace the stored IFC file path in object "
        msg += self.Object.Label + " with the new one: "
        msg += newfile
        msg += " ?"
        dlg = QtGui.QMessageBox.question(
            None,
            "Replace IFC file path?",
            msg,
            QtGui.QMessageBox.Yes | QtGui.QMessageBox.No,
            QtGui.QMessageBox.No,
        )
        if dlg == QtGui.QMessageBox.Yes:
            self.Object.IfcFilePath = newfile
            self.Object.Modified = False
            return True
        else:
            return False

    def schema_warning(self):
        from PySide import QtGui  # lazy import

        msg = "Warning: This operation will change the whole IFC file contents "
        msg += "and will not give versionable results. It is best to not do "
        msg += "this while you are in the middle of a project. "
        msg += "Continue anyway?"
        dlg = QtGui.QMessageBox.question(
            None,
            "Replace IFC file schema?",
            msg,
            QtGui.QMessageBox.Yes | QtGui.QMessageBox.No,
            QtGui.QMessageBox.No,
        )
        if dlg == QtGui.QMessageBox.Yes:
            return True
        else:
            return False

    def diff(self):
        from . import ifc_diff

        diff = ifc_diff.get_diff(self.Object)
        ifc_diff.show_diff(diff)


class ifc_vp_group:
    """View provider for the IFC group object"""

    def attach(self, vobj):
        self.Object = vobj.Object

    def getIcon(self):
        from PySide import QtGui  # lazy loading
        import Draft_rc
        import Arch_rc

        if "Layer" in self.Object.Name:
            return ":/icons/Draft_Layer.svg"
        elif "Material" in self.Object.Name:
            return ":/icons/Arch_Material_Group.svg"
        elif not hasattr(self, "modicon"):
            self.modicon = overlay(
                QtGui.QIcon.fromTheme("folder", QtGui.QIcon(":/icons/folder.svg")),
                ":/icons/IFC.svg",
            )
        return self.modicon


class ifc_vp_material:
    """View provider for the IFC group object"""

    def attach(self, vobj):
        self.Object = vobj.Object

    def getDisplayModes(self, obj):
        return []

    def getDefaultDisplayMode(self):
        return "Default"

    def setDisplayMode(self, mode):
        return mode

    def __getstate__(self):
        return None

    def __setstate__(self, state):
        return None

    def getIcon(self):
        if hasattr(self, "icondata"):
            return self.icondata
        else:
            import Arch_rc

            return ":/icons/Arch_Material.svg"

    def updateData(self, obj, prop):
        from PySide import QtCore, QtGui  # lazy loading

        if hasattr(self.Object, "Color"):
            c = self.Object.Color
            matcolor = QtGui.QColor(int(c[0] * 255), int(c[1] * 255), int(c[2] * 255))
            darkcolor = QtGui.QColor(int(c[0] * 125), int(c[1] * 125), int(c[2] * 125))
        else:
            matcolor = QtGui.QColor(200, 200, 200)
            darkcolor = QtGui.QColor(120, 120, 120)
        im = QtGui.QImage(48, 48, QtGui.QImage.Format_ARGB32)
        im.fill(QtCore.Qt.transparent)
        pt = QtGui.QPainter(im)
        pt.setPen(QtGui.QPen(QtCore.Qt.black, 2, QtCore.Qt.SolidLine, QtCore.Qt.FlatCap))
        gradient = QtGui.QLinearGradient(0, 0, 48, 48)
        gradient.setColorAt(0, matcolor)
        gradient.setColorAt(1, darkcolor)
        pt.setBrush(QtGui.QBrush(gradient))
        pt.drawEllipse(6, 6, 36, 36)
        pt.setPen(QtGui.QPen(QtCore.Qt.white, 1, QtCore.Qt.SolidLine, QtCore.Qt.FlatCap))
        pt.setBrush(QtGui.QBrush(QtCore.Qt.white, QtCore.Qt.SolidPattern))
        pt.drawEllipse(12, 12, 12, 12)
        pt.end()
        ba = QtCore.QByteArray()
        b = QtCore.QBuffer(ba)
        b.open(QtCore.QIODevice.WriteOnly)
        im.save(b, "XPM")
        self.icondata = ba.data().decode("latin1")

    def claimChildren(self):
        if hasattr(self.Object, "Group"):
            return self.Object.Group
        return []

    def setupContextMenu(self, vobj, menu):
        from . import ifc_tools  # lazy import
        from . import ifc_psets
        from PySide import QtGui  # lazy import

        if FreeCADGui.activeWorkbench().name() != "BIMWorkbench":
            return

        icon = QtGui.QIcon(":/icons/IFC.svg")
        if ifc_psets.has_psets(self.Object):
            action_props = QtGui.QAction(icon, translate("BIM", "Expand Property Sets"), menu)
            action_props.triggered.connect(self.showProps)
            menu.addAction(action_props)

    def showProps(self):
        """Expands property sets"""

        from . import ifc_psets  # lazy loading

        ifc_psets.show_psets(self.Object)
        self.Object.Document.recompute()


class ifc_vp_buildingpart(ifc_vp_object, ArchBuildingPart.ViewProviderBuildingPart):
    """A vp that inherits the Arch BuildingPart vp, but keeps aggregating properties of ifc vp"""

    def __init__(self, vobj):
        ArchBuildingPart.ViewProviderBuildingPart.__init__(self, vobj)

    def attach(self, vobj):
        ArchBuildingPart.ViewProviderBuildingPart.attach(self, vobj)


def overlay(icon1, icon2):
    """Overlays icon2 onto icon1"""

    from PySide import QtCore, QtGui  # lazy loading

    if isinstance(icon1, QtGui.QIcon):
        baseicon = icon1.pixmap(32, 32)
        baseicon = QtGui.QImage(
            baseicon.toImage().convertToFormat(QtGui.QImage.Format_ARGB32_Premultiplied)
        )
    elif isinstance(icon1, str):
        baseicon = QtGui.QImage(icon1)
    if isinstance(icon2, str):
        overlay = QtGui.QImage(icon2)
    width = baseicon.width() / 2
    overlay = overlay.scaled(width, width)
    painter = QtGui.QPainter()
    painter.begin(baseicon)
    painter.drawImage(1, 1, overlay)
    painter.end()
    ba = QtCore.QByteArray()
    b = QtCore.QBuffer(ba)
    b.open(QtCore.QIODevice.WriteOnly)
    baseicon.save(b, "XPM")
    return ba.data().decode("latin1")


def get_filepath(project):
    """Saves the associated IFC file to another file"""

    from . import ifc_tools  # lazy import
    from PySide import QtGui  # lazy import

    sf = QtGui.QFileDialog.getSaveFileName(
        None,
        "Save an IFC file",
        project.IfcFilePath,
        "Industry Foundation Classes (*.ifc)",
    )
    if sf and sf[0]:
        sf = sf[0]
        if not sf.lower().endswith(".ifc"):
            sf += ".ifc"
        project.IfcFilePath = sf
        return sf
    return None


def get_icon(vp):
    """Returns an icon for a view provider"""

    from PySide import QtCore, QtGui  # lazy import

    if hasattr(vp, "Object"):
        if hasattr(vp.Object, "IfcClass"):
            rclass = vp.Object.IfcClass.replace("StandardCase", "")
            rclass = vp.Object.IfcClass.replace("Type", "")
            ifcicon = ":/icons/IFC/" + rclass + ".svg"
            if QtCore.QFile.exists(ifcicon):
                if getattr(vp, "ifcclass", "") != rclass:
                    vp.ifcclass = rclass
                    vp.ifcicon = overlay(ifcicon, ":/icons/IFC.svg")
                return getattr(vp, "ifcicon", overlay(ifcicon, ":/icons/IFC.svg"))
            elif vp.Object.IfcClass == "IfcGroup":
                return QtGui.QIcon.fromTheme("folder", QtGui.QIcon(":/icons/folder.svg"))
            elif vp.Object.ShapeMode == "Shape":
                return ":/icons/IFC_object.svg"
            else:
                return ":/icons/IFC_mesh.svg"
    return ":/icons/IFC_object.svg"



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2017 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""The BIM ImagePlane command"""

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate


class BIM_ImagePlane:
    def GetResources(self):
        return {
            "Pixmap": "BIM_ImagePlane.svg",
            "MenuText": QT_TRANSLATE_NOOP("BIM_ImagePlane", "Image Plane"),
            "ToolTip": QT_TRANSLATE_NOOP("BIM_ImagePlane", "Creates a plane from an image"),
        }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):
        from PySide import QtGui
        import draftguitools.gui_trackers as DraftTrackers

        self.doc = FreeCAD.ActiveDocument
        self.tracker = DraftTrackers.rectangleTracker()
        self.basepoint = None
        self.opposite = None
        (filename, _filter) = QtGui.QFileDialog.getOpenFileName(
            QtGui.QApplication.activeWindow(),
            translate("BIM", "Select Image"),
            None,
            translate("BIM", "Image file (*.png *.jpg *.bmp)"),
        )
        if filename:
            FreeCAD.activeDraftCommand = self  # register as a Draft command for auto grid on/off
            self.filename = filename
            im = QtGui.QImage(self.filename)
            self.proportion = float(im.height()) / float(im.width())
            if hasattr(FreeCADGui, "Snapper"):
                FreeCADGui.Snapper.getPoint(
                    callback=self.PointCallback, movecallback=self.MoveCallback
                )

    def MoveCallback(self, point, snapinfo):
        import DraftVecUtils

        if point and self.basepoint and (point != self.basepoint):
            chord = point.sub(self.basepoint)
            length = DraftVecUtils.project(chord, self.tracker.u).Length
            height = length * self.proportion
            self.opposite = (
                FreeCAD.Vector(self.tracker.u)
                .multiply(length)
                .add(FreeCAD.Vector(self.tracker.v).multiply(height))
            )
            self.opposite = self.basepoint.add(self.opposite)
            self.tracker.update(self.opposite)

    def PointCallback(self, point, snapinfo):
        import os
        import DraftVecUtils
        import WorkingPlane

        if not point:
            # cancelled
            FreeCAD.activeDraftCommand = None
            FreeCADGui.Snapper.off()
            self.tracker.off()
            return
        elif not self.basepoint:
            # this is our first clicked point, nothing to do just yet
            self.basepoint = point
            self.tracker.setorigin(point)
            self.tracker.on()
            FreeCADGui.Snapper.getPoint(
                last=point, callback=self.PointCallback, movecallback=self.MoveCallback
            )
        else:
            # this is our second point
            FreeCAD.activeDraftCommand = None
            FreeCADGui.Snapper.off()
            self.tracker.off()
            midpoint = self.basepoint.add(self.opposite.sub(self.basepoint).multiply(0.5))
            wp = WorkingPlane.get_working_plane()
            rotation = wp.get_placement().Rotation
            diagonal = self.opposite.sub(self.basepoint)
            length = DraftVecUtils.project(diagonal, wp.u).Length
            height = DraftVecUtils.project(diagonal, wp.v).Length
            self.doc.openTransaction("Create image plane")
            image = self.doc.addObject("Image::ImagePlane", "ImagePlane")
            image.Label = os.path.splitext(os.path.basename(self.filename))[0]
            image.ImageFile = self.filename
            image.Placement = FreeCAD.Placement(midpoint, rotation)
            image.XSize = length
            image.YSize = height
            self.doc.commitTransaction()
            self.doc.recompute()


FreeCADGui.addCommand("BIM_ImagePlane", BIM_ImagePlane())



# SPDX-License-Identifier: LGPL-2.1-or-later

# (c) 2021 Werner Mayer LGPL

from PySide import QtUiTools
import FreeCADGui as Gui


class QUiLoader(QtUiTools.QUiLoader):
    """
    This is an extension of Qt's QUiLoader to also create custom widgets
    """
    def __init__(self, arg = None):
        super(QUiLoader, self).__init__(arg)
        self.ui = Gui.PySideUic

    def createWidget(self, className, parent = None, name = ""):
        widget = self.ui.createCustomWidget(className, parent, name)
        if not widget:
            widget = super(QUiLoader, self).createWidget(className, parent, name)
        return widget




# SPDX-License-Identifier: LGPL-2.1-or-later

# -*- coding: utf8 -*-
## @package importDWG
#  \ingroup DRAFT
#  \brief DWG file importer & exporter
"""
@package importDWG
ingroup DRAFT
\brief DWG file importer & exporter

This module provides support for importing and exporting Autodesk DWG files.
This module is only a thin layer that uses the ODA (formerly Teigha) File
Converter application to convert to/from DXF. Then the real work is done by
importDXF

Test files
https://knowledge.autodesk.com/support/autocad/downloads/
    caas/downloads/content/autocad-sample-files.html
"""
# Check code quality with
# flake8 --ignore=E226,E266,E401,W503

# ***************************************************************************
# *   Copyright (c) 2009 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
from FreeCAD import Console as FCC
from draftutils import params

if FreeCAD.GuiUp:
    from draftutils.translate import translate
else:

    def translate(context, txt):
        return txt


def open(filename):
    """Open filename and parse using importDXF.open().

    Parameters
    ----------
    filename : str
        The path to the filename to be opened.

    Returns
    -------
    App::Document
        The new FreeCAD document object created, with the parsed information.
    """
    dxf = convertToDxf(filename)
    if dxf:
        import importDXF

        doc = importDXF.open(dxf)
        return doc
    return


def insert(filename, docname):
    """Imports a file using importDXF.insert().

    If no document exist, it is created.

    Parameters
    ----------
    filename : str
        The path to the filename to be opened.
    docname : str
        The name of the active App::Document if one exists, or
        of the new one created.

    Returns
    -------
    App::Document
        The active FreeCAD document, or the document created if none exists,
        with the parsed information.
    """
    dxf = convertToDxf(filename)
    if dxf:
        import importDXF

        # Warning: function doesn't return?
        doc = importDXF.insert(dxf, docname)
        return doc
    return


def export(objectslist, filename):
    """Export the DWG file with a given list of objects.

    The objects are exported with importDXF.export().
    Then the result is converted to DWG.

    Parameters
    ----------
    exportList : list
        List of document objects to export.
    filename : str
        Path to the new file.

    Returns
    -------
    str
        The same `filename` input.
    """
    import importDXF
    import os
    import tempfile

    outdir = tempfile.mkdtemp()
    _basename = os.path.splitext(os.path.basename(filename))[0]
    dxf = outdir + os.sep + _basename + ".dxf"
    importDXF.export(objectslist, dxf)
    convertToDwg(dxf, filename)
    return filename


def get_libredwg_converter(typ):
    """Find the LibreDWG converter.

    It searches the FreeCAD parameters database, then searches the OS search path.
    There are no standard installation paths.

    `typ` is required because LibreDWG uses two converters and we store only one.

    Parameters
    ----------
    typ : str
        "dwg2dxf" or "dxf2dwg".

    Returns
    -------
    str
        The full path of the converter.
    """
    import os
    import platform

    path = params.get_param("TeighaFileConverter")

    if "dwg2dxf" in path or "dxf2dwg" in path:  # path set manually
        if typ not in path:
            path = os.path.dirname(path) + "/" + typ + os.path.splitext(path)[1]
        if os.path.exists(path) and os.path.isfile(path):
            return path
    elif platform.system() == "Windows":
        for sub in os.getenv("PATH").split(os.pathsep):
            path = sub.replace("\\", "/") + "/" + typ + ".exe"
            if os.path.exists(path) and os.path.isfile(path):
                return path
    else:  # for Linux and macOS
        for sub in os.getenv("PATH").split(os.pathsep):
            path = sub + "/" + typ
            if os.path.exists(path) and os.path.isfile(path):
                return path

    return None


def get_oda_converter():
    """Find the ODA converter.

    It searches the FreeCAD parameters database, then searches for common paths.

    Parameters
    ----------
    None

    Returns
    -------
    str
        The full path of the converter.
    """
    import os
    import platform

    path = params.get_param("TeighaFileConverter")

    if "ODAFileConverter" in path:  # path set manually
        if os.path.exists(path) and os.path.isfile(path):
            return path
    elif platform.system() == "Windows":
        odadir = os.path.expandvars("%ProgramFiles%\\ODA").replace("\\", "/")
        if os.path.exists(odadir):
            for sub in os.listdir(odadir):
                path = odadir + "/" + sub + "/" + "ODAFileConverter.exe"
                if os.path.exists(path) and os.path.isfile(path):
                    return path
    elif platform.system() == "Linux":
        path = "/usr/bin/ODAFileConverter"
        if os.path.exists(path) and os.path.isfile(path):
            return path
    else:  # for macOS
        path = "/Applications/ODAFileConverter.app/Contents/MacOS/ODAFileConverter"
        if os.path.exists(path) and os.path.isfile(path):
            return path

    return None


def get_qcad_converter():
    """Find the QCAD converter.

    It searches the FreeCAD parameters database, then searches for common paths.

    Parameters
    ----------
    None

    Returns
    -------
    str
        The full path of the converter.
    """
    import os
    import platform

    path = params.get_param("TeighaFileConverter")

    if "dwg2dwg" in path:  # path set manually
        pass
    elif platform.system() == "Windows":
        path = os.path.expandvars("%ProgramFiles%\\QCAD\\dwg2dwg.bat").replace("\\", "/")
    elif platform.system() == "Linux":
        # /home/$USER/opt/qcad-3.28.1-trial-linux-qt5.14-x86_64/dwg2dwg
        path = os.path.expandvars("/home/$USER/opt")
        if os.path.exists(path) and os.path.isdir(path):
            for sub in os.listdir(path):
                if "qcad" in sub:
                    path = path + "/" + sub + "/" + "dwg2dwg"
                    break
    else:  # for macOS
        path = "/Applications/QCAD.app/Contents/Resources/dwg2dwg"

    if os.path.exists(path) and os.path.isfile(path):
        return path
    return None


def convertToDxf(dwgfilename):
    """Convert a DWG file to a DXF file.

    If the converter is found it is used, otherwise the conversion fails.

    Parameters
    ----------
    dwgfilename : str
        The input filename.

    Returns
    -------
    str
        The new file produced.
    """
    import os
    import subprocess
    import tempfile

    dwgfilename = dwgfilename.replace("\\", "/")
    conv = params.get_param("DWGConversion")
    error_msg = (
        translate(
            "draft",
            """Error during DWG conversion.
Try moving the DWG file to a directory path without spaces and non-english characters,
or try saving to a lower DWG version.""",
        )
        + "\n"
    )

    if conv in [0, 1]:  # LibreDWG
        libredwg = get_libredwg_converter("dwg2dxf")
        if libredwg is not None:
            outdir = tempfile.mkdtemp().replace("\\", "/")
            basename = os.path.basename(dwgfilename)
            result = outdir + "/" + os.path.splitext(basename)[0] + ".dxf"
            cmdline = [libredwg, dwgfilename, "-o", result]
            FCC.PrintMessage(translate("draft", "Converting:") + " " + str(cmdline) + "\n")
            proc = subprocess.Popen(cmdline)
            proc.communicate()
            if os.path.exists(result):
                FCC.PrintMessage(translate("draft", "Conversion successful") + "\n")
                return result
            else:
                FCC.PrintError(error_msg)
        elif conv != 0:
            FCC.PrintError(translate("draft", "LibreDWG converter not found") + "\n")

    if conv in [0, 2]:  # ODA
        oda = get_oda_converter()
        if oda is not None:
            indir = os.path.dirname(dwgfilename)
            outdir = tempfile.mkdtemp().replace("\\", "/")
            basename = os.path.basename(dwgfilename)
            cmdline = [oda, indir, outdir, "ACAD2000", "DXF", "0", "1", basename]
            FCC.PrintMessage(translate("draft", "Converting:") + " " + str(cmdline) + "\n")
            proc = subprocess.Popen(cmdline)
            proc.communicate()
            result = outdir + "/" + os.path.splitext(basename)[0] + ".dxf"
            if os.path.exists(result):
                FCC.PrintMessage(translate("draft", "Conversion successful") + "\n")
                return result
            else:
                FCC.PrintError(error_msg)
        elif conv != 0:
            FCC.PrintError(translate("draft", "ODA converter not found") + "\n")

    if conv in [0, 3]:  # QCAD
        qcad = get_qcad_converter()
        if qcad is not None:
            outdir = tempfile.mkdtemp().replace("\\", "/")
            basename = os.path.basename(dwgfilename)
            result = outdir + "/" + os.path.splitext(basename)[0] + ".dxf"
            cmdline = [qcad, "-f", "-o", result, dwgfilename]
            FCC.PrintMessage(translate("draft", "Converting:") + " " + str(cmdline) + "\n")
            proc = subprocess.Popen(cmdline, cwd=os.path.dirname(qcad))  # cwd required for Windows
            proc.communicate()
            if os.path.exists(result):
                FCC.PrintMessage(translate("draft", "Conversion successful") + "\n")
                return result
            else:
                FCC.PrintError(error_msg)
        elif conv != 0:
            FCC.PrintError(translate("draft", "QCAD converter not found") + "\n")

    FCC.PrintError(
        translate(
            "draft",
            """No suitable external DWG converter has been found.
Please set one manually under menu Edit → Preferences → Import/Export → DWG
For more information see:
https://wiki.freecad.org/Import_Export_Preferences""",
        )
        + "\n"
    )
    return None


def convertToDwg(dxffilename, dwgfilename):
    """Convert a DXF file to a DWG file.

    If the converter is found it is used, otherwise the conversion fails.

    Parameters
    ----------
    dxffilename : str
        The input DXF file
    dwgfilename : str
        The output DWG file

    Returns
    -------
    str
        The same `dwgfilename` file path.
    """
    import os
    import subprocess

    dxffilename = dxffilename.replace("\\", "/")
    dwgfilename = dwgfilename.replace("\\", "/")
    conv = params.get_param("DWGConversion")

    if conv in [0, 1]:  # LibreDWG
        libredwg = get_libredwg_converter("dxf2dwg")
        if libredwg is not None:
            cmdline = [libredwg, dxffilename, "-y", "-o", dwgfilename]
            FCC.PrintMessage(translate("draft", "Converting:") + " " + str(cmdline) + "\n")
            proc = subprocess.Popen(cmdline)
            proc.communicate()
            return dwgfilename
        elif conv != 0:
            FCC.PrintError(translate("draft", "LibreDWG converter not found") + "\n")

    if conv in [0, 2]:  # ODA
        oda = get_oda_converter()
        if oda is not None:
            indir = os.path.dirname(dxffilename)
            outdir = os.path.dirname(dwgfilename)
            basename = os.path.basename(dxffilename)
            cmdline = [oda, indir, outdir, "ACAD2000", "DWG", "0", "1", basename]
            FCC.PrintMessage(translate("draft", "Converting:") + " " + str(cmdline) + "\n")
            proc = subprocess.Popen(cmdline)
            proc.communicate()
            return dwgfilename
        elif conv != 0:
            FCC.PrintError(translate("draft", "ODA converter not found") + "\n")

    if conv in [0, 3]:  # QCAD
        qcad = get_qcad_converter()
        if qcad is not None:
            cmdline = [qcad, "-f", "-o", dwgfilename, dxffilename]
            FCC.PrintMessage(translate("draft", "Converting:") + " " + str(cmdline) + "\n")
            proc = subprocess.Popen(cmdline, cwd=os.path.dirname(qcad))  # cwd required for Windows
            proc.communicate()
            return dwgfilename
        elif conv != 0:
            FCC.PrintError(translate("draft", "QCAD converter not found") + "\n")

    FCC.PrintError(
        translate(
            "draft",
            """No suitable external DWG converter has been found.
Please set one manually under menu Edit → Preferences → Import/Export → DWG
For more information see:
https://wiki.freecad.org/Import_Export_Preferences""",
        )
        + "\n"
    )
    return None



# test for FreeCAD issue https://github.com/freecad/freecad/issues/25893
# regression test for sketch placement updates during edit mode
# when sketcher workbench is active and has not be closed / exited

import unittest
import FreeCAD

# check if GUI is available
try:
    import FreeCADGui

    GUI_AVAILABLE = FreeCADGui.getMainWindow() is not None
except (ImportError, AttributeError):
    GUI_AVAILABLE = False

from FreeCAD import Base


class TestSketchPlacementUpdate(unittest.TestCase):
    """
    test that sketch placement/attachment changes update the 3D view
    when the sketch is in edit mode.

    regression test for issue #25893 caused by PR #25478

    NOTE: issue #25893 specifically affects ATTACHED sketches where
    AttachmentOffset changes don't update the visual position during
    edit mode. These tests use an attached sketch to properly test
    this scenario.

    These tests require GUI and will be skipped with freecadcmd.
    """

    def setUp(self):
        """create a document with a body, and within the body create a cylinder,
        then attach a sketch to the cylinder to the bottom round face of the
        cylinder.
        """
        if not GUI_AVAILABLE:
            self.skipTest("GUI not available")

        self.doc = FreeCAD.newDocument("TestPlacementUpdate")

        # create a body (typical partdesign workflow)
        self.body = self.doc.addObject("PartDesign::Body", "Body")

        # create a cylinder to attach the sketch to (10mm tall)
        self.cylinder = self.doc.addObject("PartDesign::AdditiveCylinder", "Cylinder")
        self.cylinder.Height = 10.0
        self.cylinder.Radius = 2.0
        self.body.addObject(self.cylinder)
        self.doc.recompute()

        # create a sketch and attach it to the bottom face of the cylinder
        self.sketch = self.doc.addObject("Sketcher::SketchObject", "Sketch")
        self.body.addObject(self.sketch)

        # attach the sketch to Face2, ie. bottom circle face of the cylinder
        self.sketch.AttachmentSupport = (self.cylinder, ["Face2"])
        self.sketch.MapMode = "FlatFace"
        self.sketch.MapReversed = True

        # add simple geometry so sketch is valid
        import Part

        # create 2 diagonal lines at the base of the cylinder
        self.sketch.addGeometry(
            Part.LineSegment(Base.Vector(-3, -3, 0), Base.Vector(3, 3, 0)), False
        )
        self.sketch.addGeometry(
            Part.LineSegment(Base.Vector(3, -3, 0), Base.Vector(-3, 3, 0)), False
        )

        self.doc.recompute()

    def tearDown(self):
        """clean up the test document"""
        if GUI_AVAILABLE:
            FreeCAD.closeDocument(self.doc.Name)

    @unittest.skipIf(not GUI_AVAILABLE, "GUI not available")
    def test_attachment_offset_updates_in_edit_mode(self):
        """
        test that changing AttachmentOffset while editing updates the transform.
        this is the main regression test for issue #25893.

        For attached sketches, AttachmentOffset controls the visual position
        relative to the attachment face.
        """
        # enter edit mode
        FreeCADGui.ActiveDocument.setEdit(self.sketch.Name)

        # get initial editing transform (it's a property, not a method)
        initial_transform = FreeCADGui.ActiveDocument.EditingTransform
        self.assertIsNotNone(initial_transform, "Editing transform should exist")

        # change the attachment offset (move 5mm in Z direction)
        # for attached sketches, this should update the visual position
        new_offset = Base.Placement(Base.Vector(0, 0, 5), Base.Rotation())
        self.sketch.AttachmentOffset = new_offset
        self.doc.recompute()

        # get updated editing transform
        updated_transform = FreeCADGui.ActiveDocument.EditingTransform

        # verify the transform was updated
        self.assertNotEqual(
            initial_transform,
            updated_transform,
            "Editing transform should update when AttachmentOffset changes",
        )

        # exit edit mode
        FreeCADGui.ActiveDocument.resetEdit()

    @unittest.skipIf(not GUI_AVAILABLE, "GUI not available")
    def test_multiple_attachment_offset_updates(self):
        """
        test that multiple AttachmentOffset changes in edit mode all update correctly.
        """
        # enter edit mode
        FreeCADGui.ActiveDocument.setEdit(self.sketch.Name)

        transforms = []

        # make several changes and verify each updates the transform
        for i in range(1, 4):
            z_offset = i * 2.0  # 2mm, 4mm, 6mm
            self.sketch.AttachmentOffset = Base.Placement(
                Base.Vector(0, 0, z_offset), Base.Rotation()
            )
            self.doc.recompute()

            transform = FreeCADGui.ActiveDocument.EditingTransform
            transforms.append(transform)

            # verify this transform is different from previous
            if i > 1:
                self.assertNotEqual(
                    transforms[-1],
                    transforms[-2],
                    f"Transform {i} should differ from transform {i-1}",
                )

        # exit edit mode
        FreeCADGui.ActiveDocument.resetEdit()

    @unittest.skipIf(not GUI_AVAILABLE, "GUI not available")
    def test_no_update_when_not_editing(self):
        """
        verify that attachment offset changes don't cause issues when sketch is not in edit mode.
        this ensures the fix only acts during edit mode.
        """
        # change attachment offset while NOT editing
        new_offset = Base.Placement(Base.Vector(0, 0, 8), Base.Rotation())
        self.sketch.AttachmentOffset = new_offset
        self.doc.recompute()

        # this should not crash or cause any issues
        # just verify the sketch is still valid and attached
        self.assertTrue(
            self.sketch.isDerivedFrom("Sketcher::SketchObject"), "Sketch should still be valid"
        )
        self.assertEqual(self.sketch.MapMode, "FlatFace", "Sketch should still be attached")

    @unittest.skipIf(not GUI_AVAILABLE, "GUI not available")
    def test_unattached_sketch_placement_updates(self):
        """
        test that unattached sketches also work correctly.
        for unattached sketches, the base Placement property controls position.
        """
        # create an unattached sketch for comparison
        unattached = self.doc.addObject("Sketcher::SketchObject", "UnattachedSketch")
        self.body.addObject(unattached)

        import Part

        unattached.addGeometry(Part.Circle(Base.Vector(0, 0, 0), Base.Vector(0, 0, 1), 2.0), False)
        self.doc.recompute()

        # enter edit mode
        FreeCADGui.ActiveDocument.setEdit(unattached.Name)

        initial_transform = FreeCADGui.ActiveDocument.EditingTransform

        # for unattached sketches, change Placement (not AttachmentOffset)
        unattached.Placement = Base.Placement(Base.Vector(5, 5, 5), Base.Rotation())
        self.doc.recompute()

        updated_transform = FreeCADGui.ActiveDocument.EditingTransform

        # verify transform updated
        self.assertNotEqual(
            initial_transform,
            updated_transform,
            "Editing transform should update when Placement changes for unattached sketch",
        )

        # exit edit mode
        FreeCADGui.ActiveDocument.resetEdit()



# SPDX-License-Identifier: LGPL-2.1-or-later

import unittest
from typing import Any, List, Mapping
from Path.Tool.assets import Asset, AssetUri


class TestAsset(Asset):
    asset_type: str = "test_asset"

    @classmethod
    def dependencies(cls, data: bytes) -> List[AssetUri]:
        return []

    @classmethod
    def from_bytes(cls, data: bytes, id: str, dependencies: Mapping[AssetUri, Any]) -> Any:
        return "dummy_object"

    def to_bytes(self) -> bytes:
        return b"dummy_serialized_data"

    def get_id(self) -> str:
        # Dummy implementation for testing purposes
        return "dummy_id"


class TestPathToolAsset(unittest.TestCase):
    def test_asset_cannot_be_instantiated(self):
        with self.assertRaises(TypeError):
            Asset()  # type: ignore

    def test_asset_can_be_instantiated_and_has_members(self):
        asset = TestAsset()
        self.assertIsInstance(asset, Asset)
        self.assertEqual(asset.asset_type, "test_asset")
        self.assertEqual(asset.to_bytes(), b"dummy_serialized_data")
        self.assertEqual(TestAsset.dependencies(b"some_data"), [])
        self.assertEqual(TestAsset.from_bytes(b"some_data", "some_id", {}), "dummy_object")
        self.assertEqual(asset.get_id(), "dummy_id")


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   (c) 2009, 2010                                                        *
# *   Yorik van Havre <yorik@uncreated.net>, Ken Cline <cline@frii.com>     *
# *   (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides utility functions to do operations with groups.

The functions here are also used in the Arch Workbench as some of
the objects created with this workbench work like groups.
"""
## @package groups
# \ingroup draftutils
# \brief Provides utility functions to do operations with groups.

## \addtogroup draftutils
# @{
import FreeCAD as App
import draftutils.utils as utils

from draftutils.translate import translate
from draftutils.messages import _err


def is_group(obj):
    """Return True if the given object is considered a group.

    Parameters
    ----------
    obj : App::DocumentObject
        The object to check.

    Returns
    -------
    bool
        Returns `True` if `obj` is considered a group:

        The object is derived from `App::DocumentObjectGroup` but not
        a `'LayerContainer'`.

        Or the object is of the type `'Project'`, `'Site'`, `'Building'`,
        `'Floor'`, `'BuildingPart'` or `'Space'` from the Arch Workbench.
        Note that `'Floor'` and `'Building'` are obsolete types.

        Otherwise returns `False`.
    """
    typ = utils.get_type(obj)
    return (obj.isDerivedFrom("App::DocumentObjectGroup") and typ != "LayerContainer") or typ in (
        "Project",
        "Site",
        "Building",
        "Floor",
        "BuildingPart",
        "Space",
    )


def get_group_names(doc=None):
    """Return a list of names of existing groups in the document.

    Parameters
    ----------
    doc: App::Document, optional
        It defaults to `None`.
        A document on which to search group names.
        It if is `None` it will search the current document.

    Returns
    -------
    list of str
        A list of names of objects that are considered groups.
        See the is_group function.

        Otherwise returns an empty list.
    """
    if not doc:
        found, doc = utils.find_doc(App.activeDocument())

    if not found:
        _err(translate("draft", "No active document. Aborting."))
        return []

    glist = []

    for obj in doc.Objects:
        if is_group(obj):
            glist.append(obj.Name)

    return glist


def getGroupNames():
    """Return a list of group names. DEPRECATED."""
    utils.use_instead("get_group_names")
    return get_group_names()


def ungroup(obj):
    """Remove the object from any group to which it belongs.

    A "group" is any object returned by `get_group_names`.

    Parameters
    ----------
    obj: App::DocumentObject or str
        Any type of object.
        If it is a string, it must be the `Label` of that object.
        Since a label is not guaranteed to be unique in a document,
        it will use the first object found with this label.
    """
    if isinstance(obj, str):
        obj_str = obj

    found, obj = utils.find_object(obj, doc=App.activeDocument())
    if not found:
        _err(translate("draft", "Wrong input: object {} not in document.").format(obj_str))
        return None

    doc = obj.Document

    for name in get_group_names():
        group = doc.getObject(name)
        if obj in group.Group:
            # The list of objects cannot be modified directly,
            # so a new list is created, this new list is modified,
            # and then it is assigned over the older list.
            objects = group.Group
            objects.remove(obj)
            group.Group = objects


def get_windows(obj):
    """Return the windows and rebars inside a host.

    Parameters
    ----------
    obj: App::DocumentObject
        A scripted object of type `'Wall'` or `'Structure'`
        (Arch Workbench).
        This will be searched for objects of type `'Window'` and `'Rebar'`,
        and clones of them, and the found elements will be added
        to the output list.

        The function will search recursively all elements under `obj.OutList`,
        in case the windows and rebars are nested under other walls
        and structures.

    Returns
    -------
    list
        A list of all found windows and rebars in `obj`.
        If `obj` is itself a `'Window'` or a `'Rebar'`, or a clone of them,
        it will return the same `obj` element.
    """
    out = []
    if utils.get_type(obj) in ("Wall", "Structure"):
        for o in obj.OutList:
            out.extend(get_windows(o))
        for i in obj.InList:
            if utils.get_type(i.getLinkedObject()) == "Window" or utils.is_clone(obj, "Window"):
                if hasattr(i, "Hosts"):
                    if obj in i.Hosts:
                        out.append(i)
            elif utils.get_type(i) == "Rebar" or utils.is_clone(obj, "Rebar"):
                if hasattr(i, "Host"):
                    if obj == i.Host:
                        out.append(i)
    elif utils.get_type(obj.getLinkedObject()) in ("Window", "Rebar") or utils.is_clone(
        obj, ["Window", "Rebar"]
    ):
        out.append(obj)

    return out


def get_group_contents(objectslist, walls=False, addgroups=False, spaces=False, noarchchild=False):
    """Return a list of objects from expanding the input groups.

    The function accepts any type of object, although it is most useful
    with "groups", as it is meant to unpack the objects inside these groups.

    Parameters
    ----------
    objectslist: list
        If any object in the list is considered a group, see the `is_group`
        function, its contents (`obj.Group`) are extracted and added to the
        output list.

        Single items that aren't groups are added to the output list
        as is.

    walls: bool, optional
        It defaults to `False`.
        If it is `True`, Wall and Structure objects (Arch Workbench)
        are treated as groups; they are scanned for Window, Door,
        and Rebar objects, and these are added to the output list.

    addgroups: bool, optional
        It defaults to `False`.
        If it is `True`, the group itself is kept as part of the output list.

    spaces: bool, optional
        It defaults to `False`.
        If it is `True`, Arch Spaces are added to the output list even
        when addgroups is False (their contents are always added).

    noarchchild: bool, optional
        It defaults to `False`.
        If it is `True`, the objects inside Building and BuildingParts
        (Arch Workbench) aren't added to the output list.

    Returns
    -------
    list
        The list of objects from each group present in `objectslist`,
        plus any other individual object given in `objectslist`.
    """
    newlist = []
    if not isinstance(objectslist, list):
        objectslist = [objectslist]

    for obj in objectslist:
        if obj:
            if is_group(obj):
                if addgroups or (spaces and utils.get_type(obj) == "Space"):
                    newlist.append(obj)
                if not (noarchchild and utils.get_type(obj) in ("Building", "BuildingPart")):
                    newlist.extend(
                        get_group_contents(obj.Group, walls, addgroups, spaces, noarchchild)
                    )
            else:
                # print("adding ", obj.Name)
                newlist.append(obj)
                if walls:
                    newlist.extend(get_windows(obj))

    # Clean possible duplicates
    cleanlist = []
    for obj in newlist:
        if obj not in cleanlist:
            cleanlist.append(obj)

    return cleanlist


def getGroupContents(objectslist, walls=False, addgroups=False, spaces=False, noarchchild=False):
    """Return a list of objects from groups. DEPRECATED."""
    utils.use_instead("get_group_contents")

    return get_group_contents(objectslist, walls, addgroups, spaces, noarchchild)


def get_movable_children(objectslist, recursive=True, _donelist=[]):
    """Return a list of objects with child objects that move with a host.

    Builds a list of objects with all child objects (`obj.OutList`)
    that have their `MoveWithHost` attribute set to `True`.
    This function is mostly useful for Arch Workbench objects.

    Parameters
    ----------
    objectslist: list of App::DocumentObject
        A single scripted object or list of objects.

    recursive: bool, optional
        It defaults to `True`, in which case the function
        is called recursively to also extract the children of children
        objects.
        Otherwise, only direct children of the input objects
        are added to the output list.

    _donelist: list
        List of object names. Used internally to prevent an endless loop.

    Returns
    -------
    list
        List of children objects that have their `MoveWithHost` attribute
        set to `True`.
    """
    added = []
    if not isinstance(objectslist, list):
        objectslist = [objectslist]

    for obj in objectslist:
        if obj.Name in _donelist:
            continue

        _donelist.append(obj.Name)

        # Skips some objects that should never move their children
        if utils.get_type(obj) not in (
            "App::Part",
            "PartDesign::Body",
            "Clone",
            "SectionPlane",
            "Facebinder",
            "BuildingPart",
            "App::Link",
        ):
            children = obj.OutList
            if (
                hasattr(obj, "Proxy")
                and obj.Proxy
                and hasattr(obj.Proxy, "getSiblings")
                and utils.get_type(obj) != "Window"
            ):
                # children.extend(obj.Proxy.getSiblings(obj))
                pass

            for child in children:
                if hasattr(child, "MoveWithHost") and child.MoveWithHost:
                    if hasattr(obj, "CloneOf") and obj.CloneOf:
                        if obj.CloneOf.Name != child.Name:
                            added.append(child)
                    else:
                        added.append(child)

            if recursive:
                added.extend(get_movable_children(children, recursive, _donelist))

    return added


def getMovableChildren(objectslist, recursive=True):
    """Return a list of objects with child objects. DEPRECATED."""
    utils.use_instead("get_movable_children")
    return get_movable_children(objectslist, recursive)


## @}



import FreeCAD
import unittest
from .TechDrawTestUtilities import createPageWithSVGTemplate


class DrawProjectionGroupTest(unittest.TestCase):
    def setUp(self):
        """Creates a box and sphere to make a fusions from them.
        Then creates a page and projection group"""
        FreeCAD.newDocument("TDGroup")
        FreeCAD.setActiveDocument("TDGroup")
        FreeCAD.ActiveDocument = FreeCAD.getDocument("TDGroup")
        self.document = FreeCAD.ActiveDocument
        print("document created")

        # make Fusion feature
        box = FreeCAD.ActiveDocument.addObject("Part::Box", "Box")
        box.recompute()
        print("box created")
        sphere = FreeCAD.ActiveDocument.addObject("Part::Sphere", "Sphere")
        sphere.recompute()
        print("sphere created")
        self.fusion = FreeCAD.ActiveDocument.addObject("Part::MultiFuse", "Fusion")
        FreeCAD.ActiveDocument.Fusion.Shapes = [box, sphere]
        self.fusion.recompute()
        print("Fusion created")

        # make a page
        print("making a page")
        self.page = createPageWithSVGTemplate()
        print("Page created")

    def tearDown(self):
        FreeCAD.closeDocument("TDGroup")

    def testMakeProjectionGroup(self):
        """Tests if a projection group can be added to view1"""
        # make projection group
        print("making a projection group")
        self.document.openTransaction("Create Proj Group")
        groupName = "ProjGroup"
        group = FreeCAD.ActiveDocument.addObject("TechDraw::DrawProjGroup", groupName)
        self.page.addView(group)
        print("Group created")
        group.Source = [self.fusion]

        print("adding views")
        group.addProjection("Front")  # need an Anchor
        print("added Front")

        # update group
        anchorDir = FreeCAD.Vector(0.0, 0.0, 1.0)
        anchorRot = FreeCAD.Vector(1.0, 0.0, 0.0)
        group.Anchor.Direction = anchorDir
        group.Anchor.RotationVector = anchorRot
        print("Anchor values set")
        group.Anchor.recompute()
        self.document.commitTransaction()
        print("Front/Anchor recomputed")
        group.addProjection("Left")
        print("added Left")
        group.addProjection("Top")
        print("added Top")
        group.addProjection("Right")
        print("added Right")
        group.addProjection("Rear")
        print("added Rear")
        group.addProjection("Bottom")
        print("added Bottom")

        # remove a view from projection group
        group.removeProjection("Left")
        print("removed Left")

        # test getItemByLabel method
        print("testing getItemByLabel")
        label = "Top"
        item = group.getItemByLabel(label)
        print("Item Label: " + label + " Item Name: " + item.Name)

        print("recomputing document")
        FreeCAD.ActiveDocument.recompute()

        for v in group.Views:
            print("View: " + v.Label + " " + v.TypeId)
            v.autoPosition()

        group.recompute()

        self.assertTrue("Up-to-date" in group.State)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2016 sliptonic <shopinthewoods@gmail.com>               *
# *   Copyright (c) 2025 Billy Huddleston <billy@ivdc.com>                  *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""
This file has utilities for checking and catching common errors in FreeCAD
CAM projects.  Ideally, the user could execute these utilities from an icon
to make sure tools are selected and configured and defaults have been revised
"""

from collections import Counter
from datetime import datetime
import FreeCAD
import Path
import Path.Log
import Path.Main.Sanity.ImageBuilder as ImageBuilder
import Path.Main.Sanity.ReportGenerator as ReportGenerator
import os
import Path.Dressup.Utils as PathDressup

translate = FreeCAD.Qt.translate

if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


class CAMSanity:
    # Toggle: True = use thumbnail, False = always use fallback image
    USE_TOOL_THUMBNAIL = False

    """
    This class has the functionality to harvest data from a CAM Job
    and export it in a format that is useful to the user.
    """

    def __init__(self, job, output_file):
        self.job = job
        self.output_file = output_file
        self.filelocation = os.path.dirname(output_file)

        # set the filelocation to the parent of the output filename
        if not os.path.isdir(self.filelocation):
            raise ValueError(
                translate(
                    "CAM_Sanity",
                    "output location {} doesn't exist".format(os.path.dirname(output_file)),
                )
            )

        self.image_builder = ImageBuilder.ImageBuilderFactory.get_image_builder(self.filelocation)
        self.data = self.summarize()

    def summarize(self):
        """
        Gather all the incremental parts of the analysis
        """

        data = {}
        data["baseData"] = self._baseObjectData()
        data["designData"] = self._designData()
        data["toolData"] = self._toolData()
        data["runData"] = self._runData()
        data["outputData"] = self._outputData()
        data["fixtureData"] = self._fixtureData()
        data["stockData"] = self._stockData()
        # data["squawkData"] = self._squawkData()

        return data

    def squawk(self, operator, note, date=datetime.now(), squawkType="NOTE"):
        squawkType = squawkType if squawkType in ("NOTE", "WARNING", "CAUTION", "TIP") else "NOTE"

        if squawkType == "TIP":
            squawk_icon = "Sanity_Bulb"
        elif squawkType == "NOTE":
            squawk_icon = "Sanity_Note"
        elif squawkType == "WARNING":
            squawk_icon = "Sanity_Warning"
        elif squawkType == "CAUTION":
            squawk_icon = "Sanity_Caution"

        path = f"{FreeCAD.getHomePath()}Mod/CAM/Path/Main/Sanity/{squawk_icon}.svg"

        local_date_str = date.strftime("%c")
        squawk = {
            "Date": local_date_str,
            "Operator": operator,
            "Note": note,
            "squawkType": squawkType,
            "squawkIcon": path,
        }

        return squawk

    def _baseObjectData(self):
        data = {"baseimage": "", "bases": "", "squawkData": []}
        obj = self.job
        bases = {}
        for name, count in Counter(
            [obj.Proxy.baseObject(obj, o).Label for o in obj.Model.Group]
        ).items():
            bases[name] = str(count)
            data["baseimage"] = self.image_builder.build_image(
                obj.Model, "baseimage", as_bytes=True
            )
        data["bases"] = bases

        return data

    def _designData(self):
        """
        Returns header information about the design document
        Returns information about issues and concerns (squawks)
        """

        obj = self.job
        data = {
            "FileName": "",
            "LastModifiedDate": "",
            "Customer": "",
            "Designer": "",
            "JobDescription": "",
            "JobLabel": "",
            "Sequence": "",
            "JobType": "",
            "squawkData": [],
        }
        data["FileName"] = obj.Document.FileName
        data["LastModifiedDate"] = str(obj.Document.LastModifiedDate)
        data["Customer"] = obj.Document.Company
        lastmod = obj.Document.LastModifiedDate
        if lastmod:
            try:
                # Parse ISO 8601 string and format
                data["LastModifiedDate"] = datetime.fromisoformat(str(lastmod)).strftime("%c")
            except Exception:
                data["LastModifiedDate"] = str(lastmod)
        else:
            data["LastModifiedDate"] = ""
        data["JobDescription"] = obj.Description
        data["JobLabel"] = obj.Label

        n = 0
        m = 0
        for i in obj.Document.Objects:
            if hasattr(i, "Proxy"):
                if isinstance(i.Proxy, Path.Main.Job.ObjectJob):
                    m += 1
                    if i is obj:
                        n = m
        data["Sequence"] = "{} of {}".format(n, m)
        data["JobType"] = "2.5D Milling"  # improve after job types added

        return data

    def _fixtureData(self):
        obj = self.job
        data = {"fixtures": "", "orderBy": "", "datumImage": "", "squawkData": []}

        data["fixtures"] = str(obj.Fixtures)
        data["orderBy"] = str(obj.OrderOutputBy)

        data["datumImage"] = self.image_builder.build_image(obj, "datumImage", as_bytes=True)

        return data

    def _outputData(self):
        obj = self.job
        data = {
            "lastpostprocess": "",
            "lastgcodefile": "",
            "optionalstops": "",
            "programmer": "",
            "machine": "",
            "postprocessor": "",
            "postprocessorFlags": "",
            "filesize": "",
            "linecount": "",
            "outputfilename": "setupreport",
            "squawkData": [],
        }

        data["lastpostprocess"] = str(obj.LastPostProcessDate)
        data["lastgcodefile"] = str(obj.LastPostProcessOutput)
        data["optionalstops"] = "False"
        data["programmer"] = ""
        data["machine"] = ""
        data["postprocessor"] = str(obj.PostProcessor)
        data["postprocessorFlags"] = str(obj.PostProcessorArgs)

        if obj.PostProcessorOutputFile != "":
            fname = obj.PostProcessorOutputFile
            data["outputfilename"] = os.path.splitext(os.path.basename(fname))[0]

        for op in obj.Operations.Group:
            if "Stop" in op.Name and hasattr(op, "Stop") and op.Stop is True:
                data["optionalstops"] = "True"

        if obj.LastPostProcessOutput == "":
            data["filesize"] = str(0.0)
            data["linecount"] = str(0)
            data["squawkData"].append(
                self.squawk(
                    "CAMSanity",
                    translate("CAM_Sanity", "The Job has not been post-processed"),
                )
            )
        else:
            if os.path.isfile(obj.LastPostProcessOutput):
                data["filesize"] = str(os.path.getsize(obj.LastPostProcessOutput) / 1000)
                data["linecount"] = str(sum(1 for line in open(obj.LastPostProcessOutput)))
            else:
                data["filesize"] = str(0.0)
                data["linecount"] = str(0)
                data["squawkData"].append(
                    self.squawk(
                        "CAMSanity",
                        translate(
                            "CAM_Sanity",
                            "The Job's last post-processed file is missing",
                        ),
                    )
                )

        return data

    def _runData(self):
        obj = self.job
        data = {
            "cycletotal": "",
            "jobMinZ": "",
            "jobMaxZ": "",
            "jobDescription": "",
            "operations": [],
            "squawkData": [],
        }

        data["cycletotal"] = str(obj.CycleTime)
        data["jobMinZ"] = FreeCAD.Units.Quantity(
            obj.Path.BoundBox.ZMin, FreeCAD.Units.Length
        ).UserString
        data["jobMaxZ"] = FreeCAD.Units.Quantity(
            obj.Path.BoundBox.ZMax, FreeCAD.Units.Length
        ).UserString
        data["jobDescription"] = obj.Description

        data["operations"] = []
        for op in obj.Operations.Group:
            oplabel = op.Label
            Path.Log.debug(oplabel)
            ctime = op.CycleTime if hasattr(op, "CycleTime") else "00:00:00"
            cool = op.CoolantMode if hasattr(op, "CoolantMode") else "N/A"

            o = op
            while "Dressup" in o.Name:
                oplabel = "{}:{}".format(oplabel, o.Base.Label)
                o = o.Base
                if hasattr(o, "CycleTime"):
                    ctime = o.CycleTime
                cool = o.CoolantMode if hasattr(o, "CoolantMode") else cool

            if hasattr(op, "Active") and not op.Active:
                oplabel = "{} (INACTIVE)".format(oplabel)
                ctime = "00:00:00"

            if op.Path.BoundBox.isValid():
                zmin = FreeCAD.Units.Quantity(
                    op.Path.BoundBox.ZMin, FreeCAD.Units.Length
                ).UserString
                zmax = FreeCAD.Units.Quantity(
                    op.Path.BoundBox.ZMax, FreeCAD.Units.Length
                ).UserString
            else:
                zmin = ""
                zmax = ""

            opdata = {
                "opName": oplabel,
                "minZ": zmin,
                "maxZ": zmax,
                "cycleTime": ctime,
                "coolantMode": cool,
            }
            data["operations"].append(opdata)

        return data

    def _stockData(self):
        obj = self.job
        data = {
            "xLen": "",
            "yLen": "",
            "zLen": "",
            "material": "",
            "surfaceSpeedCarbide": "",
            "surfaceSpeedHSS": "",
            "stockImage": "",
            "squawkData": [],
        }

        bb = obj.Stock.Shape.BoundBox
        data["xLen"] = FreeCAD.Units.Quantity(bb.XLength, FreeCAD.Units.Length).UserString
        data["yLen"] = FreeCAD.Units.Quantity(bb.YLength, FreeCAD.Units.Length).UserString
        data["zLen"] = FreeCAD.Units.Quantity(bb.ZLength, FreeCAD.Units.Length).UserString

        data["material"] = "Not Specified"
        if hasattr(obj.Stock, "ShapeMaterial"):
            if obj.Stock.ShapeMaterial is not None:
                data["material"] = obj.Stock.ShapeMaterial.Name

            props = obj.Stock.ShapeMaterial.PhysicalProperties
            if "SurfaceSpeedCarbide" in props:
                data["surfaceSpeedCarbide"] = FreeCAD.Units.Quantity(
                    props["SurfaceSpeedCarbide"]
                ).UserString
            if "SurfaceSpeedHSS" in props:
                data["surfaceSpeedHSS"] = FreeCAD.Units.Quantity(
                    props["SurfaceSpeedHSS"]
                ).UserString

        if data["material"] in ["Default", "Not Specified"]:
            data["squawkData"].append(
                self.squawk(
                    "CAMSanity",
                    translate("CAM_Sanity", "Consider Specifying the Stock Material"),
                    squawkType="TIP",
                )
            )

        data["stockImage"] = self.image_builder.build_image(obj.Stock, "stockImage", as_bytes=True)

        return data

    def _toolData(self):
        """
        Returns information about the tools used in the job, and associated
        toolcontrollers
        Returns information about issues and problems with the tools (squawks)
        """

        obj = self.job
        data = {"squawkData": []}

        for TC in obj.Tools.Group:
            if not hasattr(TC.Tool, "BitBody"):
                data["squawkData"].append(
                    data["squawkData"].append(
                        self.squawk(
                            "CAMSanity",
                            translate(
                                "CAM_Sanity",
                                "Tool number {} is a legacy tool. Legacy tools not \
                    supported by Path-Sanity",
                            ).format(TC.ToolNumber),
                            squawkType="WARNING",
                        )
                    )
                )
                continue  # skip old-style tools
            tooldata = data.setdefault(str(TC.ToolNumber), {})
            bitshape = tooldata.setdefault("ShapeType", "")
            if bitshape not in ["", TC.Tool.ShapeType]:
                data["squawkData"].append(
                    self.squawk(
                        "CAMSanity",
                        translate("CAM_Sanity", "Tool number {} used by multiple tools").format(
                            TC.ToolNumber
                        ),
                        squawkType="CAUTION",
                    )
                )
            tooldata["bitShape"] = TC.Tool.ShapeType
            tooldata["description"] = TC.Tool.Label
            tooldata["manufacturer"] = ""
            tooldata["url"] = ""
            tooldata["inspectionNotes"] = ""
            tooldata["diameter"] = str(TC.Tool.Diameter.UserString)
            tooldata["shape"] = TC.Tool.ShapeType

            tooldata["partNumber"] = ""

            # Use the toggle to determine which image to use
            imagebytes = None
            if self.USE_TOOL_THUMBNAIL:
                # Try to get the thumbnail
                thumb_bytes = None
                if hasattr(TC.Tool, "Proxy") and hasattr(TC.Tool.Proxy, "get_thumbnail"):
                    try:
                        thumb_bytes = TC.Tool.Proxy.get_thumbnail()
                    except Exception:
                        thumb_bytes = None
                if thumb_bytes:
                    imagebytes = thumb_bytes
                else:
                    # Warn and use fallback head-on image
                    data["squawkData"].append(
                        self.squawk(
                            "CAMSanity",
                            translate("CAM_Sanity", "Toolbit Shape for TC: {} not found").format(
                                TC.ToolNumber
                            ),
                            squawkType="WARNING",
                        )
                    )
                    imagebytes = self.image_builder.build_image(
                        TC.Tool, f"T{TC.ToolNumber}", as_bytes=True, view="headon"
                    )
            else:
                # Always use fallback head-on image
                imagebytes = self.image_builder.build_image(
                    TC.Tool, f"T{TC.ToolNumber}", as_bytes=True, view="headon"
                )
            tooldata["imagebytes"] = imagebytes
            imagepath = os.path.join(self.filelocation, f"T{TC.ToolNumber}.png")
            tooldata["imagepath"] = imagepath
            Path.Log.debug(imagepath)
            # No longer writing imagedata to disk; handled by imagebytes logic above

            tooldata["feedrate"] = str(TC.HorizFeed.UserString)
            if TC.HorizFeed.Value == 0.0:
                data["squawkData"].append(
                    self.squawk(
                        "CAMSanity",
                        translate("CAM_Sanity", "Tool Controller '{}' has no feedrate").format(
                            TC.Label
                        ),
                        squawkType="WARNING",
                    )
                )

            tooldata["spindlespeed"] = f"{int(TC.SpindleSpeed)} rpm"
            if TC.SpindleSpeed == 0.0:
                data["squawkData"].append(
                    self.squawk(
                        "CAMSanity",
                        translate("CAM_Sanity", "Tool Controller '{}' has no spindlespeed").format(
                            TC.Label
                        ),
                        squawkType="WARNING",
                    )
                )

            used = False
            for op in obj.Operations.Group:
                base_op = PathDressup.baseOp(op)
                if hasattr(base_op, "ToolController") and base_op.ToolController is TC:
                    used = True
                    tooldata.setdefault("ops", []).append(
                        {
                            "Operation": base_op.Label,
                            "ToolController": TC.Label,
                            "Feed": str(TC.HorizFeed.UserString),
                            "Speed": f"{int(TC.SpindleSpeed)} rpm",
                        }
                    )

            if used is False:
                tooldata.setdefault("ops", [])
                data["squawkData"].append(
                    self.squawk(
                        "CAMSanity",
                        translate("CAM_Sanity", "Tool Controller '{}' is not used").format(
                            TC.Label
                        ),
                        squawkType="WARNING",
                    )
                )

        return data

    def serialize(self, obj):
        """A function to serialize non-serializable objects."""
        if isinstance(obj, type(Exception)):
            # Convert an exception to its string representation
            return str(obj)
        # You might need to handle more types depending on your needs
        return str(obj)  # Fallback to convert any other non-serializable types to string

    def get_output_report(self):
        Path.Log.debug("get_output_url")

        generator = ReportGenerator.ReportGenerator(self.data, embed_images=True)
        html = generator.generate_html()
        generator = None
        return html



# ***************************************************************************
# *   Copyright (c) 2021 Werner Mayer <werner.wm.mayer@gmx.de>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# ***************************************************************************/


import math, os, sys, unittest
import FreeCAD
import Part
import Sketcher

App = FreeCAD


class TestSketchExpression(unittest.TestCase):
    def setUp(self):
        self.Doc = FreeCAD.newDocument("TestSketchExpr")

    # See https://forum.freecad.org/viewtopic.php?f=3&t=64699
    # and https://forum.freecad.org/viewtopic.php?f=10&t=64718
    def testConstraintWithExpression(self):
        sketch = self.Doc.addObject("Sketcher::SketchObject", "Sketch")

        geoList = []
        geoList.append(Part.LineSegment(App.Vector(0, 5, 0), App.Vector(5, 5, 0)))
        geoList.append(Part.LineSegment(App.Vector(5, 5, 0), App.Vector(5, 0, 0)))
        geoList.append(Part.LineSegment(App.Vector(5, 0, 0), App.Vector(0, 0, 0)))
        geoList.append(Part.LineSegment(App.Vector(0, 0, 0), App.Vector(0, 5, 0)))
        sketch.addGeometry(geoList, False)

        conList = []
        conList.append(Sketcher.Constraint("Coincident", 0, 2, 1, 1))
        conList.append(Sketcher.Constraint("Coincident", 1, 2, 2, 1))
        conList.append(Sketcher.Constraint("Coincident", 2, 2, 3, 1))
        conList.append(Sketcher.Constraint("Coincident", 3, 2, 0, 1))
        conList.append(Sketcher.Constraint("Horizontal", 0))
        conList.append(Sketcher.Constraint("Horizontal", 2))
        conList.append(Sketcher.Constraint("Vertical", 1))
        conList.append(Sketcher.Constraint("Vertical", 3))
        sketch.addConstraint(conList)
        del geoList, conList

        # Check that circular references get detected
        length = sketch.addConstraint(Sketcher.Constraint("Distance", 0, 6.0))
        height = sketch.addConstraint(Sketcher.Constraint("Distance", 3, 5.0))
        sketch.renameConstraint(length, "Length")
        sketch.renameConstraint(height, "Height")

        sketch.setExpression("Constraints[{}]".format(height), ".Constraints.Length")

        with self.assertRaises(RuntimeError) as context:
            sketch.setExpression("Constraints.Length", ".Constraints.Length * 2")
        assert ".Constraints.Length reference creates a cyclic dependency." in str(
            context.exception
        )

        with self.assertRaises(RuntimeError) as context:
            sketch.setExpression("Constraints.Length", ".Constraints.Height * 2")
        assert ".Constraints.Length reference creates a cyclic dependency." in str(
            context.exception
        )

    def tearDown(self):
        # comment out to omit closing document for debugging
        FreeCAD.closeDocument(self.Doc.Name)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022 Mario Passaglia                                    *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************


"""Provides the viewprovider code for the Shapestring object."""

import FreeCADGui as Gui

from draftviewproviders.view_base import ViewProviderDraft
from drafttaskpanels.task_shapestring import ShapeStringTaskPanelEdit


class ViewProviderShapeString(ViewProviderDraft):

    def __init__(self, vobj):

        vobj.Proxy = self

    def getIcon(self):

        return ":/icons/Draft_ShapeString.svg"

    def setEdit(self, vobj, mode):
        if mode != 0:
            return None

        if (
            not "Draft_Edit" in Gui.listCommands()
        ):  # Using Draft_Edit to detect if the Draft, Arch or BIM WB has been loaded.
            self.wb_before_edit = Gui.activeWorkbench()
            Gui.activateWorkbench("DraftWorkbench")
        self.task = ShapeStringTaskPanelEdit(vobj)
        Gui.Control.showDialog(self.task)
        return True

    def unsetEdit(self, vobj, mode):
        if mode != 0:
            return None

        self.task.finish()
        if hasattr(self, "wb_before_edit"):
            Gui.activateWorkbench(self.wb_before_edit.name())
            delattr(self, "wb_before_edit")
        return True



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2020 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""Provides GUI tools to set up default styles."""

## @package gui_setstyle
# \ingroup draftguitools
# \brief Provides GUI tools to set Draft styles such as color or line width

## \addtogroup draftguitools
# @{

import os
from PySide import QtCore
from PySide import QtGui
from PySide import QtWidgets

import FreeCAD as App
import FreeCADGui as Gui
import Draft_rc
from FreeCAD import Units as U
from draftutils import params
from draftutils import utils


def QT_TRANSLATE_NOOP(ctx, txt):
    return txt


translate = App.Qt.translate

__title__ = "FreeCAD Draft Workbench GUI Tools - Styling tools"
__author__ = "Yorik van Havre"
__url__ = "https://www.freecad.org"

PRESETPATH = os.path.join(App.getUserAppDataDir(), "Draft", "StylePresets.json")


class Draft_SetStyle:
    """The Draft_SetStyle FreeCAD command definition."""

    def GetResources(self):

        return {
            "Pixmap": "Draft_Apply",
            "Accel": "S, S",
            "MenuText": QT_TRANSLATE_NOOP("Draft_SetStyle", "Set Style"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_SetStyle", "Sets the default style and can apply the style to objects"
            ),
        }

    def Activated(self):

        Gui.Control.showDialog(Draft_SetStyle_TaskPanel())


class Draft_SetStyle_TaskPanel:
    """The task panel for the Draft_SetStyle command"""

    def __init__(self):

        self.form = Gui.PySideUic.loadUi(":/ui/TaskPanel_SetStyle.ui")
        self.form.setWindowIcon(
            QtGui.QIcon.fromTheme("gtk-apply", QtGui.QIcon(":/icons/Draft_Apply.svg"))
        )

        self.form.saveButton.setIcon(
            QtGui.QIcon.fromTheme("gtk-save", QtGui.QIcon(":/icons/document-save.svg"))
        )
        self.form.applyButton.setIcon(
            QtGui.QIcon.fromTheme("gtk-apply", QtGui.QIcon(":/icons/Draft_Apply.svg"))
        )
        self.form.annotButton.setIcon(QtGui.QIcon(":/icons/Draft_Text.svg"))

        self.form.ShapeColor.setProperty(
            "color", self.getColor(params.get_param_view("DefaultShapeColor"))
        )
        self.form.AmbientColor.setProperty(
            "color", self.getColor(params.get_param_view("DefaultAmbientColor"))
        )
        self.form.EmissiveColor.setProperty(
            "color", self.getColor(params.get_param_view("DefaultEmissiveColor"))
        )
        self.form.SpecularColor.setProperty(
            "color", self.getColor(params.get_param_view("DefaultSpecularColor"))
        )
        self.form.Transparency.setValue(params.get_param_view("DefaultShapeTransparency"))
        self.form.Shininess.setValue(params.get_param_view("DefaultShapeShininess"))

        self.form.LineColor.setProperty(
            "color", self.getColor(params.get_param_view("DefaultShapeLineColor"))
        )
        self.form.LineWidth.setValue(params.get_param_view("DefaultShapeLineWidth"))
        self.form.PointColor.setProperty(
            "color", self.getColor(params.get_param_view("DefaultShapeVertexColor"))
        )
        self.form.PointSize.setValue(params.get_param_view("DefaultShapePointSize"))
        self.form.DrawStyle.setCurrentIndex(params.get_param("DefaultDrawStyle"))
        self.form.DisplayMode.setCurrentIndex(params.get_param("DefaultDisplayMode"))

        self.form.TextColor.setProperty(
            "color", self.getColor(params.get_param("DefaultTextColor"))
        )
        self.form.TextFont.setCurrentFont(QtGui.QFont(params.get_param("textfont")))
        self.form.TextSize.setText(U.Quantity(params.get_param("textheight"), U.Length).UserString)
        self.form.LineSpacing.setValue(params.get_param("LineSpacing"))
        self.form.ScaleMultiplier.setValue(params.get_param("DefaultAnnoScaleMultiplier"))

        self.form.AnnoLineColor.setProperty(
            "color", self.getColor(params.get_param("DefaultAnnoLineColor"))
        )
        self.form.AnnoLineWidth.setValue(params.get_param("DefaultAnnoLineWidth"))
        self.form.ArrowStyleStart.setCurrentIndex(params.get_param("dimsymbolstart"))
        self.form.ArrowSizeStart.setText(
            U.Quantity(params.get_param("arrowsizestart"), U.Length).UserString
        )
        self.form.ArrowStyleEnd.setCurrentIndex(params.get_param("dimsymbolend"))
        self.form.ArrowSizeEnd.setText(
            U.Quantity(params.get_param("arrowsizeend"), U.Length).UserString
        )
        self.form.ShowUnit.setChecked(params.get_param("showUnit"))
        self.form.UnitOverride.setText(params.get_param("overrideUnit"))
        self.form.DimOvershoot.setText(
            U.Quantity(params.get_param("dimovershoot"), U.Length).UserString
        )
        self.form.ExtLines.setText(U.Quantity(params.get_param("extlines"), U.Length).UserString)
        self.form.ExtOvershoot.setText(
            U.Quantity(params.get_param("extovershoot"), U.Length).UserString
        )
        self.form.TextSpacing.setText(
            U.Quantity(params.get_param("dimspacing"), U.Length).UserString
        )

        self.form.saveButton.clicked.connect(self.onSaveStyle)
        self.form.applyButton.clicked.connect(self.onApplyStyle)
        self.form.annotButton.clicked.connect(self.onApplyAnnot)
        self.form.comboPresets.currentIndexChanged.connect(self.onLoadStyle)

        self.loadDefaults()

    def loadDefaults(self):

        presets = [self.form.comboPresets.itemText(0)]
        self.form.comboPresets.clear()
        pdict = self.load()
        pdict_keys = list(pdict)
        presets.extend(pdict_keys)
        self.form.comboPresets.addItems(presets)
        current = self.getValues()
        for name, preset in pdict.items():
            if all(item in current.items() for item in preset.items()):  # if preset == current:
                self.form.comboPresets.setCurrentIndex(1 + (pdict_keys.index(name)))
                break

    def getColor(self, c):

        return QtGui.QColor(utils.rgba_to_argb(c))

    def getValues(self):

        return {
            "ShapeColor": utils.argb_to_rgba(self.form.ShapeColor.property("color").rgba()),
            "AmbientColor": utils.argb_to_rgba(self.form.AmbientColor.property("color").rgba()),
            "EmissiveColor": utils.argb_to_rgba(self.form.EmissiveColor.property("color").rgba()),
            "SpecularColor": utils.argb_to_rgba(self.form.SpecularColor.property("color").rgba()),
            "Transparency": self.form.Transparency.value(),
            "Shininess": self.form.Shininess.value(),
            "LineColor": utils.argb_to_rgba(self.form.LineColor.property("color").rgba()),
            "LineWidth": self.form.LineWidth.value(),
            "PointColor": utils.argb_to_rgba(self.form.PointColor.property("color").rgba()),
            "PointSize": self.form.PointSize.value(),
            "DrawStyle": self.form.DrawStyle.currentIndex(),
            "DisplayMode": self.form.DisplayMode.currentIndex(),
            "TextColor": utils.argb_to_rgba(self.form.TextColor.property("color").rgba()),
            "TextFont": self.form.TextFont.currentFont().family(),
            "TextSize": U.Quantity(self.form.TextSize.text()).Value,
            "LineSpacing": self.form.LineSpacing.value(),
            "ScaleMultiplier": self.form.ScaleMultiplier.value(),
            "AnnoLineColor": utils.argb_to_rgba(self.form.AnnoLineColor.property("color").rgba()),
            "AnnoLineWidth": self.form.AnnoLineWidth.value(),
            "ArrowStyleStart": self.form.ArrowStyleStart.currentIndex(),
            "ArrowSizeStart": U.Quantity(self.form.ArrowSizeStart.text()).Value,
            "ArrowStyleEnd": self.form.ArrowStyleEnd.currentIndex(),
            "ArrowSizeEnd": U.Quantity(self.form.ArrowSizeEnd.text()).Value,
            "ShowUnit": self.form.ShowUnit.isChecked(),
            "UnitOverride": self.form.UnitOverride.text(),
            "DimOvershoot": U.Quantity(self.form.DimOvershoot.text()).Value,
            "ExtLines": U.Quantity(self.form.ExtLines.text()).Value,
            "ExtOvershoot": U.Quantity(self.form.ExtOvershoot.text()).Value,
            "TextSpacing": U.Quantity(self.form.TextSpacing.text()).Value,
        }

    def setValues(self, preset):

        # For compatibility with <= V0.21 some properties, if missing, reference others:
        #     'new' prop    -> old prop
        #     ---------------------------
        #     PointColor    -> LineColor
        #     PointSize     -> LineWidth
        #     AnnoLineColor -> LineColor
        #     AnnoLineWidth -> LineWidth

        def getDefDraft(entry):
            return params.get_param(entry, ret_default=True)

        def getDefView(entry):
            return params.get_param_view(entry, ret_default=True)

        # ---------------------------------------------------------------------

        self.form.ShapeColor.setProperty(
            "color", self.getColor(preset.get("ShapeColor", getDefView("DefaultShapeColor")))
        )
        self.form.AmbientColor.setProperty(
            "color", self.getColor(preset.get("AmbientColor", getDefView("DefaultAmbientColor")))
        )
        self.form.EmissiveColor.setProperty(
            "color", self.getColor(preset.get("EmissiveColor", getDefView("DefaultEmissiveColor")))
        )
        self.form.SpecularColor.setProperty(
            "color", self.getColor(preset.get("SpecularColor", getDefView("DefaultSpecularColor")))
        )
        self.form.Transparency.setValue(
            preset.get("Transparency", getDefView("DefaultShapeTransparency"))
        )
        self.form.Shininess.setValue(preset.get("Shininess", getDefView("DefaultShapeShininess")))

        # ---------------------------------------------------------------------

        self.form.LineColor.setProperty(
            "color", self.getColor(preset.get("LineColor", getDefView("DefaultShapeLineColor")))
        )
        self.form.LineWidth.setValue(preset.get("LineWidth", getDefView("DefaultShapeLineWidth")))
        self.form.PointColor.setProperty(
            "color",
            self.getColor(
                preset.get(
                    "PointColor", preset.get("LineColor", getDefView("DefaultShapeVertexColor"))
                )
            ),
        )
        self.form.PointSize.setValue(
            preset.get("PointSize", preset.get("LineWidth", getDefView("DefaultShapePointSize")))
        )
        self.form.DrawStyle.setCurrentIndex(
            preset.get("DrawStyle", getDefDraft("DefaultDrawStyle"))
        )
        self.form.DisplayMode.setCurrentIndex(
            preset.get("DisplayMode", getDefDraft("DefaultDisplayMode"))
        )

        # ---------------------------------------------------------------------

        self.form.TextColor.setProperty(
            "color", self.getColor(preset.get("TextColor", getDefDraft("DefaultTextColor")))
        )
        self.form.TextFont.setCurrentFont(
            QtGui.QFont(preset.get("TextFont", getDefDraft("textfont")))
        )
        self.form.TextSize.setText(
            U.Quantity(preset.get("TextSize", getDefDraft("textheight")), U.Length).UserString
        )
        self.form.LineSpacing.setValue(preset.get("LineSpacing", getDefDraft("LineSpacing")))
        self.form.ScaleMultiplier.setValue(
            preset.get("ScaleMultiplier", getDefDraft("DefaultAnnoScaleMultiplier"))
        )

        # ---------------------------------------------------------------------

        self.form.AnnoLineColor.setProperty(
            "color",
            self.getColor(
                preset.get(
                    "AnnoLineColor", preset.get("LineColor", getDefDraft("DefaultAnnoLineColor"))
                )
            ),
        )
        self.form.AnnoLineWidth.setValue(
            preset.get(
                "AnnoLineWidth", preset.get("LineWidth", getDefDraft("DefaultAnnoLineWidth"))
            )
        )
        self.form.ArrowStyleStart.setCurrentIndex(
            preset.get("ArrowStyleStart", getDefDraft("dimsymbolstart"))
        )
        self.form.ArrowSizeStart.setText(
            U.Quantity(
                preset.get("ArrowSizeStart", getDefDraft("arrowsizestart")), U.Length
            ).UserString
        )
        self.form.ArrowStyleEnd.setCurrentIndex(
            preset.get("ArrowStyleEnd", getDefDraft("dimsymbolend"))
        )
        self.form.ArrowSizeEnd.setText(
            U.Quantity(preset.get("ArrowSizeEnd", getDefDraft("arrowsizeend")), U.Length).UserString
        )
        self.form.ShowUnit.setChecked(preset.get("ShowUnit", getDefDraft("showUnit")))
        self.form.UnitOverride.setText(preset.get("UnitOverride", getDefDraft("overrideUnit")))
        self.form.DimOvershoot.setText(
            U.Quantity(preset.get("DimOvershoot", getDefDraft("dimovershoot")), U.Length).UserString
        )
        self.form.ExtLines.setText(
            U.Quantity(preset.get("ExtLines", getDefDraft("extlines")), U.Length).UserString
        )
        self.form.ExtOvershoot.setText(
            U.Quantity(preset.get("ExtOvershoot", getDefDraft("extovershoot")), U.Length).UserString
        )
        self.form.TextSpacing.setText(
            U.Quantity(preset.get("TextSpacing", getDefDraft("dimspacing")), U.Length).UserString
        )

    def reject(self):

        Gui.Control.closeDialog()

    def accept(self):

        params.set_param_view(
            "DefaultShapeColor", utils.argb_to_rgba(self.form.ShapeColor.property("color").rgba())
        )
        params.set_param_view(
            "DefaultAmbientColor",
            utils.argb_to_rgba(self.form.AmbientColor.property("color").rgba()),
        )
        params.set_param_view(
            "DefaultEmissiveColor",
            utils.argb_to_rgba(self.form.EmissiveColor.property("color").rgba()),
        )
        params.set_param_view(
            "DefaultSpecularColor",
            utils.argb_to_rgba(self.form.SpecularColor.property("color").rgba()),
        )
        params.set_param_view("DefaultShapeTransparency", self.form.Transparency.value())
        params.set_param_view("DefaultShapeShininess", self.form.Shininess.value())

        params.set_param_view(
            "DefaultShapeLineColor",
            utils.argb_to_rgba(self.form.LineColor.property("color").rgba()),
        )
        params.set_param_view("DefaultShapeLineWidth", self.form.LineWidth.value())
        params.set_param_view(
            "DefaultShapeVertexColor",
            utils.argb_to_rgba(self.form.PointColor.property("color").rgba()),
        )
        params.set_param_view("DefaultShapePointSize", self.form.PointSize.value())
        params.set_param("DefaultDrawStyle", self.form.DrawStyle.currentIndex())
        params.set_param("DefaultDisplayMode", self.form.DisplayMode.currentIndex())

        params.set_param(
            "DefaultTextColor", utils.argb_to_rgba(self.form.TextColor.property("color").rgba())
        )
        params.set_param("textfont", self.form.TextFont.currentFont().family())
        params.set_param("textheight", U.Quantity(self.form.TextSize.text()).Value)
        params.set_param("LineSpacing", self.form.LineSpacing.value())
        params.set_param("DefaultAnnoScaleMultiplier", self.form.ScaleMultiplier.value())

        params.set_param(
            "DefaultAnnoLineColor",
            utils.argb_to_rgba(self.form.AnnoLineColor.property("color").rgba()),
        )
        params.set_param("DefaultAnnoLineWidth", self.form.AnnoLineWidth.value())
        params.set_param("dimsymbolstart", self.form.ArrowStyleStart.currentIndex())
        params.set_param("arrowsizestart", U.Quantity(self.form.ArrowSizeStart.text()).Value)
        params.set_param("dimsymbolend", self.form.ArrowStyleEnd.currentIndex())
        params.set_param("arrowsizeend", U.Quantity(self.form.ArrowSizeEnd.text()).Value)
        params.set_param("showUnit", self.form.ShowUnit.isChecked())
        params.set_param("overrideUnit", self.form.UnitOverride.text())
        params.set_param("dimovershoot", U.Quantity(self.form.DimOvershoot.text()).Value)
        params.set_param("extlines", U.Quantity(self.form.ExtLines.text()).Value)
        params.set_param("extovershoot", U.Quantity(self.form.ExtOvershoot.text()).Value)
        params.set_param("dimspacing", U.Quantity(self.form.TextSpacing.text()).Value)

        self.reject()

    def onApplyStyle(self):

        for obj in Gui.Selection.getSelection():
            self.apply_style_to_obj(obj)

    def onApplyAnnot(self):

        if App.ActiveDocument is not None:  # Command can be called without a document.
            objs = App.ActiveDocument.Objects
            typs = [
                "Dimension",
                "LinearDimension",
                "AngularDimension",
                "Text",
                "DraftText",
                "Label",
            ]
            for obj in objs:
                if utils.get_type(obj) in typs:
                    self.apply_style_to_obj(obj)

    def apply_style_to_obj(self, obj):

        vobj = obj.ViewObject
        if not vobj:
            return

        properties = vobj.PropertiesList
        if "FontName" not in properties:  # Shapes
            if "ShapeAppearance" in properties:
                material = App.Material()
                material.DiffuseColor = self.form.ShapeColor.property("color").getRgbF()[
                    :3
                ]  # Remove Alpha (which is 1 instead of 0).
                material.AmbientColor = self.form.AmbientColor.property("color").getRgbF()[:3]
                material.EmissiveColor = self.form.EmissiveColor.property("color").getRgbF()[:3]
                material.SpecularColor = self.form.SpecularColor.property("color").getRgbF()[:3]
                material.Transparency = self.form.Transparency.value() / 100
                material.Shininess = self.form.Shininess.value() / 100
                vobj.ShapeAppearance = (material,)
            if "LineColor" in properties:
                vobj.LineColor = self.form.LineColor.property("color").getRgbF()[:3]
            if "LineWidth" in properties:
                vobj.LineWidth = self.form.LineWidth.value()
            if "PointColor" in properties:
                vobj.PointColor = self.form.PointColor.property("color").getRgbF()[:3]
            if "PointSize" in properties:
                vobj.PointSize = self.form.PointSize.value()
            if "DrawStyle" in properties:
                vobj.DrawStyle = utils.DRAW_STYLES[self.form.DrawStyle.currentIndex()]
            if "DisplayMode" in properties:
                dm = utils.DISPLAY_MODES[self.form.DisplayMode.currentIndex()]
                if dm in vobj.listDisplayModes():
                    vobj.DisplayMode = dm
        else:  # Annotations
            if "TextColor" in properties:
                vobj.TextColor = self.form.TextColor.property("color").getRgbF()[:3]
            if "FontName" in properties:
                vobj.FontName = self.form.TextFont.currentFont().family()
            if "FontSize" in properties:
                vobj.FontSize = U.Quantity(self.form.TextSize.text()).Value
            if "LineSpacing" in properties:
                vobj.LineSpacing = self.form.LineSpacing.value()
            if "ScaleMultiplier" in properties:
                vobj.ScaleMultiplier = self.form.ScaleMultiplier.value()
            if "LineColor" in properties:
                vobj.LineColor = self.form.AnnoLineColor.property("color").getRgbF()[:3]
            if "LineWidth" in properties:
                vobj.LineWidth = self.form.AnnoLineWidth.value()
            if "ArrowTypeStart" in properties:
                vobj.ArrowTypeStart = utils.ARROW_TYPES[self.form.ArrowStyleStart.currentIndex()]
            if "ArrowSizeStart" in properties:
                vobj.ArrowSizeStart = U.Quantity(self.form.ArrowSizeStart.text()).Value
            if "ArrowTypeEnd" in properties:
                vobj.ArrowTypeEnd = utils.ARROW_TYPES[self.form.ArrowStyleEnd.currentIndex()]
            if "ArrowSizeEnd" in properties:
                vobj.ArrowSizeEnd = U.Quantity(self.form.ArrowSizeEnd.text()).Value
            if "ShowUnit" in properties:
                vobj.ShowUnit = self.form.ShowUnit.isChecked()
            if "UnitOverride" in properties:
                vobj.UnitOverride = self.form.UnitOverride.text()
            if "DimOvershoot" in properties:
                vobj.DimOvershoot = U.Quantity(self.form.DimOvershoot.text()).Value
            if "ExtLines" in properties:
                vobj.ExtLines = U.Quantity(self.form.ExtLines.text()).Value
            if "ExtOvershoot" in properties:
                vobj.ExtOvershoot = U.Quantity(self.form.ExtOvershoot.text()).Value
            if "TextSpacing" in properties:
                vobj.TextSpacing = U.Quantity(self.form.TextSpacing.text()).Value

    def onLoadStyle(self, index):

        if index > 0:
            pdict = self.load()
            if self.form.comboPresets.itemText(index) in pdict:
                preset = pdict[self.form.comboPresets.itemText(index)]
                self.setValues(preset)

    def onSaveStyle(self):

        reply = QtWidgets.QInputDialog.getText(
            None, translate("Draft", "Save style"), translate("Draft", "Name of this new style")
        )
        if reply[1]:
            name = reply[0]
            pdict = self.load()
            if pdict:
                if name in pdict:
                    reply = QtWidgets.QMessageBox.question(
                        None,
                        translate("Draft", "Warning"),
                        translate("Draft", "Name exists. Overwrite?"),
                        QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
                        QtWidgets.QMessageBox.No,
                    )
                    if reply == QtWidgets.QMessageBox.No:
                        return
            preset = self.getValues()
            pdict[name] = preset
            self.save(pdict)
            self.loadDefaults()

    def load(self):
        """loads the presets json file, returns a dict"""

        pdict = {}
        try:
            import json
            from json.decoder import JSONDecodeError
        except Exception:
            App.Console.PrintError(
                translate("Draft", "Error: json module not found. Unable to load style") + "\n"
            )
            return
        if os.path.exists(PRESETPATH):
            with open(PRESETPATH, "r") as f:
                try:
                    pdict = json.load(f)
                except JSONDecodeError:
                    return {}
        return pdict

    def save(self, d):
        """saves the given dict to the presets json file"""

        try:
            import json
        except Exception:
            App.Console.PrintError(
                translate("Draft", "Error: json module not found. Unable to save style") + "\n"
            )
            return
        folder = os.path.dirname(PRESETPATH)
        if not os.path.exists(folder):
            os.makedirs(folder)
        with open(PRESETPATH, "w") as f:
            json.dump(d, f, indent=4)


Gui.addCommand("Draft_SetStyle", Draft_SetStyle())

## @}



#! python
# (c) 2011 Adrian Przekwas LGPL

from __future__ import division  # allows floating point division from integers
import FreeCAD, Part
from FreeCAD import Base


class MySpring:
    def __init__(self, obj):
        """Add the properties: Pitch, Diameter, Height, BarDiameter"""
        obj.addProperty(
            "App::PropertyLength", "Pitch", "MySpring", "Pitch of the helix", locked=True
        ).Pitch = 5.0
        obj.addProperty(
            "App::PropertyLength", "Diameter", "MySpring", "Diameter of the helix", locked=True
        ).Diameter = 6.0
        obj.addProperty(
            "App::PropertyLength", "Height", "MySpring", "Height of the helix", locked=True
        ).Height = 30.0
        obj.addProperty(
            "App::PropertyLength", "BarDiameter", "MySpring", "Diameter of the bar", locked=True
        ).BarDiameter = 3.0
        obj.Proxy = self

    def onChanged(self, fp, prop):
        if prop == "Pitch" or prop == "Diameter" or prop == "Height" or prop == "BarDiameter":
            self.execute(fp)

    def execute(self, fp):
        pitch = fp.Pitch
        radius = fp.Diameter / 2
        height = fp.Height
        barradius = fp.BarDiameter / 2
        myhelix = Part.makeHelix(pitch, height, radius)
        g = myhelix.Edges[0].Curve
        c = Part.Circle()
        c.Center = g.value(0)  # start point of the helix
        c.Axis = (0, 1, 0)
        c.Radius = barradius
        p = c.toShape()
        section = Part.Wire([p])
        makeSolid = True
        isFrenet = True
        myspring = Part.Wire(myhelix).makePipeShell([section], makeSolid, isFrenet)
        fp.Shape = myspring


def makeMySpring():
    doc = FreeCAD.activeDocument()
    if doc is None:
        doc = FreeCAD.newDocument()
    spring = doc.addObject("Part::FeaturePython", "My_Spring")
    spring.Label = "My Spring"
    MySpring(spring)
    spring.ViewObject.Proxy = 0
    doc.recompute()


if __name__ == "__main__":
    makeMySpring()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2019 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides GUI tools to create WorkingPlaneProxy objects."""
## @package gui_planeproxy
# \ingroup draftguitools
# \brief Provides GUI tools to create WorkingPlaneProxy objects.

## \addtogroup draftguitools
# @{
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import FreeCADGui as Gui
from draftutils import gui_utils

__title__ = "FreeCAD Draft Workbench GUI Tools - Working plane-related tools"
__author__ = "Yorik van Havre, Werner Mayer, Martin Burbaum, Ken Cline, " "Dmitry Chigrin"
__url__ = "https://www.freecad.org"


class Draft_WorkingPlaneProxy:
    """The Draft_WorkingPlaneProxy command definition."""

    def GetResources(self):
        """Set icon, menu and tooltip."""
        return {
            "Pixmap": "Draft_PlaneProxy",
            "MenuText": QT_TRANSLATE_NOOP("Draft_WorkingPlaneProxy", "Working Plane Proxy"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_WorkingPlaneProxy",
                "Creates a proxy object from the current working plane that allows to restore the camera position and visibility of objects",
            ),
        }

    def IsActive(self):
        """Return True when this command should be available."""
        return bool(gui_utils.get_3d_view())

    def Activated(self):
        """Execute when the command is called."""
        App.ActiveDocument.openTransaction("Create Working Plane Proxy")
        Gui.addModule("Draft")
        Gui.addModule("WorkingPlane")
        Gui.doCommand("pl = WorkingPlane.get_working_plane().get_placement()")
        Gui.doCommand("Draft.make_workingplaneproxy(pl)")
        App.ActiveDocument.commitTransaction()
        App.ActiveDocument.recompute()


Gui.addCommand("Draft_WorkingPlaneProxy", Draft_WorkingPlaneProxy())

## @}
